# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Dict, List

from ..lex import Lexer, LexMode, c_like_punctuation_patterns, whitespace_patterns


patterns=dict(
  # These are ordered roughly to improve matching speed on a sample (all .py files in the repository).

  # Python keywords must come before `name`.
  const = r'None|True|False|Ellipsis',
  kw_as       = 'as',
  kw_async    = 'async',
  kw_await    = 'await',
  kw_break    = 'break',
  kw_class    = 'class',
  kw_continue = 'continue',
  kw_def      = 'def',
  kw_elif     = 'elif',
  kw_else     = 'else',
  kw_for      = 'for',
  kw_from     = 'from',
  kw_if       = 'if',
  kw_import   = 'import',
  kw_while    = 'while',
  kw_yield    = 'yield',

  name  = r'[_A-Za-z][_A-Za-z0-9]*', # Most common.
  **whitespace_patterns,
  int_h = r'0x[_0-9A-Fa-f]+',
  int_b = r'0b[_01]+',
  int_o = r'0o[_07]+',
  flt   = r'([0-9]+\.|\.[0-9])[_0-9]*',
  int_d = r'[0-9][_0-9]*',

  **c_like_punctuation_patterns,

  comment_type_ignore = r'\# type: ignore',
  comment_type      = r'\# type:[\n]*',
  comment           = r'\#[^\n]*',
)

main_pattern_names = list(patterns.keys())


str_pattern_names:Dict[str,List[str]] = {}

def add_str_patterns(quote:str, label:str, multiline:bool):
  '''
  Note about lexing string literals:
  general pattern for quoting with escapes is Q([^EQ]|EQ|EE)*Q.
  It is crucial that the escape character E is excluded in the '[^EQ]' clause,
  or else when matching against 'QEQQ', the pattern greedily matches 'QEQ'.
  To allow a trailing escape character, the 'EE' clause is also required.
  '''
  q = quote
  q0 = q[0]
  n = r'\n' # Normally, newline is part of the exclusion set.
  or_ml = '' # Multiline has an additional choice clause.
  if multiline:
    n = '' # For multiline, newline is not excluded.
    or_ml = r'|' + q0 + r'{1,2}(?!' + q0 + ')' # Accept one or two quotes, so long as they aren't followed by a third.
  # The leading "(?:r[bf]|[bf]?r?)" pattern is for all the valid combinations of [rbf] flags.
  patterns.update({
    f'str_{label}'      : fr'(?:r[bf]|[bf]?r?){q}(?s:[^{n}\\{q0}]|\\.{or_ml})*{q}',
    f'str_{label}_o'    : fr'(?:r[bf]|[bf]?r?){q}',
    f'str_{label}_c'    : q,
    f'str_{label}_re'   : r'[][(){}|^?*+]+',
    f'str_{label}_txt'  : fr'[^\\{q[0]}]',
    f'str_{label}_esc'  : r'\\.',
  })


for l, q in [('s', "'"), ('d', '"')]:
  for multiline in [True, False]:
    label = l + '3' if multiline else l
    quote = q * 3 if multiline else q
    add_str_patterns(label=label, quote=quote, multiline=multiline)
    main_pattern_names.append('str_' + label) # TODO: generalize to allow choosing multimode lexer.

lexer = Lexer(flags='x', patterns=patterns, modes=[LexMode('main', main_pattern_names)])



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import inspect
import sys
from code import InteractiveInterpreter
from typing import ContextManager, List

from .io import errSL
from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


class Interpreter(InteractiveInterpreter):
  '''
  Similar to code.InteractiveConsole.
  '''

  def __init__(self, locals=None, filename='<interact>') -> None:
    super().__init__(locals=locals)
    self.buffer:List[str] = []
    self.filename = filename


  def interact(self, banner='Entering interactive loop...', exit_msg='Exited interactive loop.') -> None:
    try: sys.ps1
    except AttributeError: sys.ps1 = '> '
    try: sys.ps2
    except AttributeError: sys.ps2 = '… '
    self.write(f'{banner}\n')
    needs_more_input = 0
    while 1:
      prompt = sys.ps2 if needs_more_input else sys.ps1
      try:
        line = self.raw_input(prompt)
      except EOFError:
        self.write('\n')
        break
      except KeyboardInterrupt:
        self.write('\nKeyboardInterrupt\n')
        self.buffer.clear()
        needs_more_input = 0
      else:
        needs_more_input = self.push(line)
    if exit_msg is None:
      self.write('now exiting %s...\n' % self.__class__.__name__)
    elif exit_msg != '':
      self.write('%s\n' % exit_msg)

  def push(self, line:str) -> bool:
    '''
    Push a line to the interpreter and process if it is complete.
    '''
    self.buffer.append(line)
    source = '\n'.join(self.buffer)
    needs_more_input = self.runsource(source, self.filename)
    if not needs_more_input:
      self.buffer.clear()
      return False
    return True

  def raw_input(self, prompt:str) -> str:
    '''Write a prompt and read a line.

    The returned line does not include the trailing newline.
    When the user enters the EOF key sequence, EOFError is raised.

    The base implementation uses the built-in function
    input(); a subclass may replace this with a different
    implementation.

    '''
    return input(prompt)


def interact(banner=None, locals=None):
  if locals is None:
    stack = inspect.stack()
    parent_frame_info = stack[1]
    locals = parent_frame_info.frame.f_locals
  errSL('locals:')
  for k, v in locals.items():
    rv = repr(v)
    max_width = 128
    if len(rv) > max_width:
      rv = rv[:max_width-1] + '…'
    errSL(f'  {k:<16}: {rv}')
  interpreter = Interpreter(locals=locals)
  interpreter.interact()


class ExitOnKeyboardInterrupt(ContextManager):

  def __init__(self, dbg:bool=False) -> None:
    self.dbg = dbg

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    if isinstance(exc_value, KeyboardInterrupt) and not self.dbg:
      exit(' Keyboard interrupt.')



from zstandard import ZstdCompressor # type: ignore
from typing import AnyStr, BinaryIO, IO, Iterable, List, NoReturn, TextIO
from .typing import OptTypeBaseExc, OptBaseExc, OptTraceback


class ZstWriterBase(IO[AnyStr]):

  def __init__(self, file:BinaryIO, level=6, threads=1, chunk_size=32768) -> None:
    self.file = file
    self.compressor = ZstdCompressor(level=level, threads=1)
    self.chunker = self.compressor.chunker(chunk_size=chunk_size)
    # TODO: use stream_writer instead of chunker?
    #stream_writer = cctx.stream_writer(fh)
    self.input_byte_count = 0
    self.compressed_byte_count = 0

  def __iter__(self) -> NoReturn: raise TypeError

  def __next__(self) -> NoReturn: raise TypeError

  def fileno(self) -> int: return self.file.fileno()

  def isatty(self) -> bool: return self.file.isatty()

  def readable(self) -> bool: return self.file.readable()

  def writable(self) -> bool: return self.file.writable()

  def tell(self) -> int: return self.file.tell()

  def truncate(self, size:int=None) -> int: return self.file.truncate()

  def read(self, size=-1) -> AnyStr: raise TypeError

  def readline(self, size=-1) -> AnyStr: raise TypeError

  def readlines(self, size=-1) -> List[AnyStr]: raise TypeError

  def seek(self, offset:int, whence=0) -> int: return self.file.seek(offset, whence)

  def seekable(self) -> bool: return self.file.seekable()


  def writelines(self, lines:Iterable) -> None:
    for line in lines: self.write(line)

  def _write_chunks(self, chunks:Iterable[bytes]) -> None:
    for chunk in chunks:
      self.compressed_byte_count += self.file.write(chunk)

  def flush(self) -> None:
    self._write_chunks(self.chunker.flush())

  def close(self) -> None:
    self._write_chunks(self.chunker.finish())
    self.file.close()

  @property
  def closed(self) -> bool:
    return self.file.closed

  @property
  def compression_ratio(self) -> float:
    return self.compressed_byte_count / self.input_byte_count


class ZstWriter(ZstWriterBase, BinaryIO):

  def __enter__(self) -> 'ZstWriter': return self

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self.close()

  def write(self, data:bytes) -> int:
    l = len(data)
    self.input_byte_count += l
    self._write_chunks(self.chunker.compress(data))
    return l


class ZstTextWriter(ZstWriterBase, TextIO):

  def __init__(self, file:BinaryIO, encoding:str=None, errors:str=None, level=6, threads=1, chunk_size=32768) -> None:
    self._encoding = encoding or 'utf-8'
    self._errors = errors or 'strict'
    super().__init__(file=file, level=level, threads=threads, chunk_size=chunk_size)

  def __enter__(self) -> 'ZstTextWriter': return self

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self.close()

  def write(self, text:str) -> int:
    data = text.encode(self.encoding)
    self.input_byte_count += len(data)
    self._write_chunks(self.chunker.compress(data))
    return len(text)

  @property
  def encoding(self) -> str:
    return self._encoding

  @property
  def errors(self) -> str:
    return self._errors



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re as _re
from itertools import zip_longest as _zip_longest
from os import PathLike, fspath as _fspath, sep
from os.path import (abspath as _abspath, basename as _basename, commonpath as _commonpath, dirname as _dirname,
  isabs as _isabs, join as _join, realpath as _realpath, relpath as _relpath, split as _split, splitext as _splitext)
from typing import List, Tuple, Union


Path = Union[str, PathLike]
PathOrFd = Union[Path, int]


class MixedAbsoluteAndRelativePathsError(Exception): pass
class NotAPathError(Exception): pass
class PathIsNotDescendantError(Exception): pass


def abs_or_norm_path(path: Path, make_abs: bool) -> str:
  'Return the absolute path if make_abs is True. If make_abs is False, return a normalized path.'
  return abs_path(path) if make_abs else norm_path(path)


def abs_path(path: Path) -> str:
  'Return the absolute path corresponding to `path`.'
  return _abspath(norm_path(path))


def current_dir() -> str: return abs_path('.')


def executable_dir() -> str:
  'Return the parent directory of this excutable.'
  return _dirname(executable_path())


def executable_path() -> str:
  'Return the path to this executable.'
  import __main__ # type: ignore # mypy bug.
  path: str = __main__.__file__
  return _realpath(path)


def insert_path_stem_suffix(path: Path, suffix: str) -> str:
  'Insert a suffix in between the path stem and ext.'
  stem, ext = split_stem_ext(path)
  return f'{stem}{suffix}{ext}'


def is_norm_path(path:Path) -> bool:
  return bool(_norm_path_re.fullmatch(str_path(path)))


_norm_path_re = _re.compile(r'''(?x)
# This is nasty. In particular the "component" clauses have to deal with the validity of three or more dots.
  [./] # Either a lone dot or slash, or...
| (?: # Absolute path.
    / # Leading slash.
    (?:[^./]|\.{1,2}[^./]|\.{3})[^/]* # Component.
  )+
| # Relative path.
  (?: \.\./ )* # Leading backups ending with a slash.
  (?:
    \.\. # Final backup.
  | (?:[^./]|\.{1,2}[^./]|\.{3})[^/]* # Initial component.
    (?: / (?:[^./]|\.{1,2}[^./]|\.{3})[^/]* )* # Trailing components.
  )
''')


def is_path_abs(path: Path) -> bool:
  'Return true if `path` is an absolute path.'
  return _isabs(path)


def is_sub_path(path: Path) -> bool:
  'Return true if `path` is a relative path that, after normalization, does not refer to parent directories.'
  return not is_path_abs(path) and '..' not in path_split(path)



def norm_path(path:Path) -> str:
  '''
  Normalize `path`.
  * trailing slashes are removed.
  * duplicate slashes (empty components) and '.' components are removed.
  * '..' components are simplified, or dropped if it implies a location beyond the root '/'.
  * Unlike `os.path.normpath`, this implementation simplifies leading double slash to a single slash.
  '''
  p = str_path(path)
  lead_slash = '/' if p.startswith('/') else ''
  comps:list = []
  for comp in p.split('/'):
    if not comp or comp == '.': continue
    if comp == '..':
      if comps:
        if comps[-1] == '..': comps.append(comp)
        else: comps.pop()
      else:
        if lead_slash: pass # '..' beyond root gets dropped.
        else: comps.append(comp)
    else:
      comps.append(comp)
  return (lead_slash + '/'.join(comps)) or '.'


def parent_dir() -> str: return abs_path('..')


def path_common_prefix(*paths: Path) -> str:
  'Return the common path prefix for a sequence of paths.'
  try: return _commonpath([str_path(p) for p in paths]) # type: ignore
  except ValueError: # we want a more specific exception.
    raise MixedAbsoluteAndRelativePathsError(paths) from None


def path_compound_ext(path: Path) -> str:
  return ''.join(path_exts(path))


def path_descendants(start_path: Path, end_path: Path, include_start=True, include_end=True) -> Tuple[str, ...]:
  '''
  Return a tuple of paths from `start_path` to `end_path`.
  By default, `include_start` and `include_end` are both True.
  TODO: normalize paths, and deal with '..' case.
  '''
  prefix = path_split(start_path)
  comps = path_split(end_path)
  if not prefix: raise NotAPathError(start_path)
  if not comps: raise NotAPathError(end_path)
  if prefix == comps:
    if include_start or include_end:
      return (str_path(start_path),)
    return ()
  if prefix != comps[:len(prefix)]:
    raise PathIsNotDescendantError(end_path, start_path)
  start_i = len(prefix) + (1 if include_start else 0)
  end_i = len(comps) + (1 if include_end else 0)
  return tuple(path_join(*comps[:i]) for i in range(start_i, end_i))


def path_dir(path: Path) -> str:
  "Return the dir portion of `path` (possibly empty), e.g. 'dir/name'."
  return _dirname(str_path(path))


def path_dir_or_dot(path: Path) -> str:
  "Return the dir portion of a path, e.g. 'dir/name', or '.' in the case of no path."
  return path_dir(path) or '.'


def path_ext(path: Path) -> str:
  'The file extension of the path.'
  return split_stem_ext(path)[1]


def path_exts(path: Path) -> Tuple[str, ...]:
  exts = []
  while True:
    path, ext = split_stem_ext(path)
    if not ext: break
    exts.append(ext)
  return tuple(exts)


def path_join(first: Path, *additional: Path) -> str:
  'Join the path with the system path separator.'
  return _join(str_path(first), *[str_path(p) for p in additional])


def path_name(path: Path) -> str:
  "Return the name portion of a path (possibly including an extension); the 'basename' in Unix terminology."
  return _basename(str_path(path))


def path_name_stem(path: Path) -> str:
  'The file name without extension; the name stem will not span directories.'
  return path_stem(path_name(path))


def path_rel_to_dir(path:Path, dir:Path) -> str:
  comps:List[str] = []
  parent_comps = 0
  for p, r in _zip_longest(path_split(abs_path(path)), path_split(abs_path(dir))):
    if not parent_comps and p == r: continue
    if p is not None: comps.append(p)
    if r is not None: parent_comps += 1
  comps = ['..']*parent_comps + comps
  if not comps: return '.'
  return path_join(*comps)


def path_rel_to_ancestor(path: Path, ancestor: str, dot=False) -> str:
  '''
  Return the path relative to `ancestor`.
  If `path` is not descended from `ancestor`, raise PathIsNotDescendantError.
  If `path` and `ancestor` are equivalent (path component-wise),
   then return '.' if dot is True, or else raise PathIsNotDescendantError.
  '''
  comps = path_split(path)
  prefix = path_split(ancestor)
  if comps == prefix:
    if dot: return '.'
    raise PathIsNotDescendantError(path, ancestor)
  if prefix == comps[:len(prefix)]:
    return path_join(*comps[len(prefix):])
  raise PathIsNotDescendantError(path, ancestor)


def path_rel_to_ancestor_or_abs(path: Path, ancestor: str, dot=False) -> str:
  '''
  Return the path relative to `ancestor` if `path` is a descendant,
  or else the corresponding absolute path.
  `dot` has the same effect as in `path_rel_to_ancestor`.
  '''
  ap = abs_path(path)
  aa = abs_path(ancestor)
  try:
    return path_rel_to_ancestor(ap, aa, dot=dot)
  except PathIsNotDescendantError:
    return ap


def path_rel_to_current_or_abs(path: Path, dot=False) -> str:
  return path_rel_to_ancestor_or_abs(path, current_dir(), dot=dot)


def path_split(path: Path) -> List[str]:
  # TODO: rename to path_comps?
  np = norm_path(path)
  if np == '/': return ['/']
  assert not np.endswith('/')
  return [comp or '/' for comp in np.split(sep)]


def path_stem(path: Path) -> str:
  'The path without the file extension; the stem may span multiple directories.'
  return split_stem_ext(path)[0]


def rel_path(path: Path, start: Path='.') -> str:
  'Return a version of `path` relative to `start`, which defaults to the current directory.'
  return _relpath(str_path(path), str_path(start))


def replace_first_dir(path: Path, replacement: str) -> str:
  parts = path_split(path)
  if not parts: raise Exception('replace_first_dir: path is empty')
  parts[0] = replacement
  return path_join(*parts)


def split_dir_name(path: Path) -> Tuple[str, str]:
  "Split the path into dir and name (possibly including an extension) components, e.g. 'dir/name'."
  return _split(str_path(path))


def split_dir_stem_ext(path: Path) -> Tuple[str, str, str]:
  'Split the path into a (dir, stem, ext) triple.'
  dir, name = split_dir_name(path)
  stem, ext = split_stem_ext(name)
  return dir, stem, ext


def split_stem_ext(path: Path) -> Tuple[str, str]:
  '''
  Split the path into stem (possibly spanning directories) and extension components, e.g. 'stem.ext'.
  '''
  return _splitext(str_path(path))


def str_path(path: Path) -> str:
  p = _fspath(path)
  if isinstance(p, str): return p
  assert isinstance(p, bytes)
  return p.decode()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'Parse Python format strings and generate corresponding regular expressions.'

import re
from .string import line_col_1
from typing import Any, Iterable, Match, Pattern, Tuple


class FormatError(Exception): pass


fmt_re = re.compile(r'''(?x)
(?P<formatter>\{
        (?P<name> [^{}!:]* )
  (?: ! (?P<conv> [ars] ) )?
  (?: : (?P<spec> (?: [^{}] | \{ [^{}]* \} )* ) )?
  #^ for the spec, we allow a single level of nested formatters.
\})
| \{\{
| \}\}
| [^{}]+
''')

# Translated from standard docs "6.1.3.1. Format Specification Mini-Language".
fmt_spec_re = re.compile(r'''(?x)
(?: (?P<fill> . )? (?P<align> [<>=^]) )?
(?P<sign> [-+\ ] )?
(?P<alt> \# )?
(?P<zero> 0 ) ?
(?P<width> \d+ | \{ [^{}]* \} )? # note: nested format.
(?P<grouping> [_,] )?
(?: \. (?P<precision> \d+ | \{ [^{}]* \} ) )? # note: nested format.
(?P<type> [bcdeEfFgGnosxX%] )?
''')


spec_type_patterns = {
  'd': r'\d'
}

spec_types = {
  'd': int
}


def has_formatter(string: str) -> bool:
  'Returns True if `string` contains a format pattern.'
  for match in gen_format_matches(string):
    if match.group('formatter'):
      return True
  return False


def count_formatters(fmt: str) -> int:
  count = 0
  for match in gen_format_matches(fmt):
    if match.group('formatter'):
      count += 1
  return count


def parse_formatters(fmt: str) -> Iterable[Tuple[str, str, str, type]]:
  for match in gen_format_matches(fmt):
    formatter = match.group('formatter')
    if formatter is not None:
      value_type: type = str
      name, conv, spec = match.group('name', 'conv', 'spec')
      assert isinstance(name, str), name
      if spec:
        spec_match = fmt_spec_re.fullmatch(spec)
        if not spec_match: raise _exc(fmt, match.start(), f'invalid format spec: {spec!r}')
        fill, align, sign, alt, zero, width, grouping, precision, type_ = spec_match.group(
          'fill', 'align', 'sign', 'alt', 'zero', 'width', 'grouping', 'precision', 'type')
        if type_:
          try: value_type = spec_types[type_]
          except KeyError as e: raise _exc(fmt, match.start(), f'spec type {type_!r} not implemented') from e
      yield (name, conv or '', spec or '', value_type)


def format_partial(fmt: str, *args: str, **kwargs: Any) -> str:
  args_it = iter(args)
  def format_frag(match: Match[str]) -> str:
    formatter = match.group('formatter')
    if formatter:
      name = match.group('name')
      if name:
        try: return formatter.format(**kwargs)
        except KeyError: return formatter
      else:
        try: return formatter.format(next(args_it), **kwargs)
        except (StopIteration, KeyError): return formatter
    return match.group()
  return ''.join(format_frag(m) for m in gen_format_matches(fmt))


def format_to_re(fmt: str, allow_empty=False, greedy=False) -> Pattern[str]:
  'translate a format string into a regular expression pattern.'
  quantifier = ('*' if allow_empty else '+') + ('' if greedy else '?')

  def pattern_from(match: Match[str]) -> str:

    def exc(msg: str) -> FormatError: return _exc(fmt, match.start(), msg)

    if match.group('formatter'):
      pat = '.' + quantifier # Default pattern.
      spec = match.group('spec')
      if spec:
        spec_match = fmt_spec_re.fullmatch(spec)
        if not spec_match: raise exc(f'invalid format spec: {spec!r}')

        fill, align, sign, alt, zero, width, grouping, precision, type_ = spec_match.group(
          'fill', 'align', 'sign', 'alt', 'zero', 'width', 'grouping', 'precision', 'type')

        if type_:
          try: pat = spec_type_patterns[type_] + '+'
          except KeyError as e: raise exc(f'spec type {type_!r} not implemented') from e

      name = match.group('name')
      if name: return f'(?P<{name}>{pat})'
      else: return f'({pat})'

    text = match.group()
    if text == '{{': return '\{'
    if text == '}}': return '\}'
    return re.escape(text)

  return re.compile(''.join(pattern_from(m) for m in gen_format_matches(fmt)))


def gen_format_matches(fmt: str) -> Iterable[Match]:
  'Generate a sequence of match objects completely covering the format string.'
  pos = 0
  def exc() -> FormatError: return _exc(fmt, pos, f'invalid format character: {fmt[pos]!r}')
  for match in fmt_re.finditer(fmt):
    if match.start() != pos: raise exc()
    pos = match.end()
    yield match
  if pos != len(fmt): raise exc()


def _exc(fmt: str, pos: int, msg: str) -> FormatError:
  line, col = line_col_1(fmt, pos)
  return FormatError(f'<str>:{line}:{col}: {msg}')




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
SVG writer.
SVG elements reference: https://developer.mozilla.org/en-US/docs/Web/SVG/Element.
'''

from functools import reduce
from html import escape as html_escape
from math import floor, log10
from typing import (Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union, overload)

from .iterable import iter_unique, window_iter
from .range import Num, NumRange
from .xml.escape import EscapedStr, esc_xml_attr, esc_xml_text
from .xml.tree import IndexCounters, XmlAttrs, XmlWriter, add_opt_attrs


Dim = Union[int, float, str]
Vec = Tuple[Num, Num]
VecOrNum = Union[Vec, Num]
F2 = Tuple[float, float]
F2OrF = Union[F2, float]
BoundsF2 = Tuple[F2, F2]
PathCommand = Tuple

PointTransform = Callable[[Tuple], F2]

def _axis_transform_dummy(val:Num) -> float: raise Exception('missing transform')

TickFmt = Callable[[float], Any]


class SvgWriter(XmlWriter):

  replaced_attrs = {
    'href': 'xlink:href', # safari Version 11.1.1 requires this, even though xlink is deprecated in svg 2 standard.
    'w': 'width',
    'h': 'height',
    **XmlWriter.replaced_attrs,
  }

  def __init__(self, *children:Any, tag:str=None, title:Any=None, _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    'A `title` attribute gets converted into a child element, which renders in browsers as a tooltip.'
    super().__init__(*children, tag=tag, _counters=_counters, attrs=attrs, **kwargs)
    if title is not None:
      self.add(SvgTitle(title))


  def on_close(self) -> None:
    self.children.sort(key=lambda el: el.attrs.get('z_index', 0))


# Html and Svg share these classes.

class HtmlSvgWriter(XmlWriter):
  can_auto_close_tags = False # HTML5 dictates that each tag type either be self-closing or not.

class Script(XmlWriter):
  tag = 'script'

class Style(XmlWriter):
  tag = 'style'


# SVG leaf elements.

class Circle(SvgWriter):
  tag = 'circle'

  def __init__(self, pos:Vec=None, r:Num=None, x:Num=None, y:Num=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    add_opt_attrs(kwargs, cx=fmt_num(x), cy=fmt_num(y), r=fmt_num(r))
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Image(SvgWriter):
  tag = 'image'

  def __init__(self, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if size is not None:
      assert w is None
      assert h is None
      w, h = unpack_VecOrNum(size)
    add_opt_attrs(kwargs, x=fmt_num(x), y=fmt_num(y), width=fmt_num(w), height=fmt_num(h))
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Line(SvgWriter):
  tag = 'line'

  def __init__(self, a:Vec=None, b:Vec=None, *, x1:Num=None, y1:Num=None, x2:Num=None, y2:Num=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if a is not None:
      assert x1 is None
      assert y1 is None
      x1, y1 = a
    if b is not None:
      assert x2 is None
      assert y2 is None
      x2, y2 = b
    add_opt_attrs(kwargs, x1=fmt_num(x1), y1=fmt_num(y1), x2=fmt_num(x2), y2=fmt_num(y2))
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Path(SvgWriter):
  tag = 'path'

  def __init__(self, commands:Iterable[PathCommand],
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    assert (attrs is None or 'd' not in attrs)
    assert 'd' not in kwargs
    cmd_strs:List[str] = []
    for c in commands:
      try: code = c[0]
      except IndexError: continue
      try: exp_len = _path_command_lens[code]
      except KeyError as e: raise Exception(f'bad path command code: {c!r}') from e
      if len(c) != exp_len + 1: raise Exception(f'path command requires {exp_len} arguments: {c}')
      cmd_strs.append(code + ','.join(fmt_num(n) for n in c[1:]))
    kwargs['d'] = ' '.join(cmd_strs)
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Poly(SvgWriter):

  def __init__(self, points:Iterable[Vec],
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    assert (attrs is None or 'points' not in attrs)
    assert 'points' not in kwargs
    point_strs:List[str] = []
    for p in points:
      if len(p) < 2: raise Exception(f'bad point for {self.tag}: {p}')
      point_strs.append(f'{fmt_num(p[0])},{fmt_num(p[1])}')
    kwargs['points'] = ' '.join(point_strs)
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Polygon(Poly):
  tag = 'polygon'

class Polyline(Poly):
  tag = 'polyline'


class Rect(SvgWriter):
  tag = 'rect'

  def __init__(self, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None, r:VecOrNum=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if size is not None:
      assert w is None
      assert h is None
      w, h = unpack_VecOrNum(size)
    rx:Optional[Num]
    ry:Optional[Num]
    if isinstance(r, tuple):
      rx, ry = r
    else:
      rx = ry = r
    add_opt_attrs(kwargs, x=fmt_num(x), y=fmt_num(y), w=fmt_num(w), h=fmt_num(h), rx=fmt_num(rx), ry=fmt_num(ry))
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


class Text(SvgWriter):
  tag = 'text'

  def __init__(self, *text, pos:Vec=None, x:Num=None, y:Num=None, alignment_baseline:str=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if alignment_baseline is not None and alignment_baseline not in alignment_baselines: raise ValueError(alignment_baseline)
    add_opt_attrs(kwargs, x=fmt_num(x), y=fmt_num(y), alignment_baseline=alignment_baseline)
    super().__init__(*text, _counters=_counters, attrs=attrs, **kwargs)


class TSpan(SvgWriter):
  tag = 'tspan'



class Use(SvgWriter):
  tag = 'use'

  def __init__(self, id:str, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    assert id
    if id[0] != '#': id = '#' + id
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if size is not None:
      assert w is None
      assert h is None
      w, h = unpack_VecOrNum(size)
    add_opt_attrs(kwargs, href=id, x=fmt_num(x), y=fmt_num(y), width=fmt_num(w), height=fmt_num(h))
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)




class SvgBranch(SvgWriter):


  def circle(self, pos:Vec=None, r:Num=None, x:Num=None, y:Num=None, **attrs:Any) -> Circle:
    'Create a child `circle` element.'
    return self.child(Circle, pos=pos, r=r, x=x, y=y, **attrs)


  def clipPath(self, **attrs:Any) -> 'ClipPath':
    'Create a child `clipPath` element.'
    return self.child(ClipPath, **attrs)


  def defs(self, **attrs:Any) -> 'Defs':
    'Create a child `defs` element.'
    return self.child(Defs, **attrs)


  def g(self, transform:Iterable[str]='', **attrs:Any) -> 'G':
    'Create an SVG `g` element for use in a context manager.'
    return self.child(G, transform=transform, **attrs)


  def image(self, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None, **attrs:Any) -> Image:
    'Create a child `defs` element.'
    return self.child(Image, pos=pos, size=size, x=x, y=y, w=w, h=h, **attrs)


  def line(self, a:Vec=None, b:Vec=None, *, x1:Num=None, y1:Num=None, x2:Num=None, y2:Num=None, **attrs:Any) -> Line:
    'Create a child `defs` element.'
    return self.child(Line, a=a, b=b, x1=x1, y1=y1, x2=x2, y2=y2, **attrs)


  def marker(self, *children:Any, id:str='', pos:Vec=None, size:VecOrNum=None, x:Num=None, y:Num=None, w:Num=None, h:Num=None,
   vx:Num=0, vy:Num=0, vw:Num=None, vh:Num=None, markerUnits='strokeWidth', orient:str='auto', **attrs:Any) -> 'Marker':
    'Create a child `marker` element.'
    return self.child(Marker, *children, id=id, pos=pos, size=size, x=x, y=y, w=w, h=h, vx=vx, vy=vy, vw=vw, vh=vh,
      markerUnits=markerUnits, orient=orient, **attrs)


  def path(self, commands:Iterable[PathCommand], **attrs:Any) -> Path:
    'Create a child `path` element.'
    return self.child(Path, commands=commands, **attrs)


  def polygon(self, points:Iterable[Vec], **attrs:Any) -> Polygon:
    return self.child(Polygon, points=points, **attrs)


  def polyline(self, points:Iterable[Vec], **attrs:Any) -> Polyline:
    return self.child(Polyline, points=points, **attrs)


  def rect(self, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None, r:VecOrNum=None, **attrs:Any) -> Rect:
    return self.child(Rect, pos=pos, size=size, x=x, y=y, w=w, h=h, r=r, **attrs)


  def script(self, *text:str, **attrs,) -> Script:
    'Create a child `script` element.'
    return self.child(Script, *text, **attrs)


  def style(self, *text:str, **attrs,) -> Style:
    'Create a child `style` element.'
    return self.child(Style, *text, **attrs)


  def symbol(self, *children:Any, id:str, vx:Num=None, vy:Num=None, vw:Num=None, vh:Num=None, **attrs:Any) -> 'Symbol':
    'Create a child `symbol` element.'
    return self.child(Symbol, *children, id=id, vx=vx, vy=vy, vw=vw, vh=vh, **attrs)


  def text(self, *text:Any, pos:Vec=None, x:Num=None, y:Num=None, alignment_baseline:str=None, **attrs:Any) -> Text:
    'Create a child `text` element.'
    return self.child(Text, *text, pos=pos, x=x, y=y, alignment_baseline=alignment_baseline, **attrs)


  def use(self, id:str, pos:Vec=None, size:VecOrNum=None, *, x:Num=None, y:Num=None, w:Num=None, h:Num=None, **attrs:Any) -> Use:
    'Use a previously defined symbol'
    return self.child(Use, id=id, pos=pos, size=size, x=x, y=y, w=w, h=h, **attrs)


  # High level.

  def grid(self, pos:Vec=(0,0), size:VecOrNum=(256,256), *,
   step:VecOrNum=16, off:Vec=(0, 0), corner_radius:VecOrNum=None, **attrs:Any) -> 'G':
    # TODO: per-axis log-transform option.
    x, y = f2_for_vec(pos)
    w, h = unpack_VecOrNum(size)
    sx, sy = unpack_VecOrNum(step)
    off_x, off_y = f2_for_vec(off)
    if off_x <= 0: off_x = sx # Do not emit line at 0, because that is handled by border.
    if off_y <= 0: off_y = sy # Do not emit line at 0, because that is handled by border.
    x_start = x + off_x
    y_start = y + off_y
    x_end = x + w
    y_end = y + h
    cl = attrs.setdefault('cl', 'grid')
    # TODO: if we are really going to support rounded corners then the border rect should clip the interior lines.
    g = self.g(**attrs)
    with g:
      for tick in NumRange(x_start, x_end, sx): g.line((tick, y), (tick, y_end)) # Vertical lines.
      for tick in NumRange(y_start, y_end, sy): g.line((x, tick), (x_end, tick)) # Horizontal lines.
      g.rect(cl=cl+'-border', x=x, y=y, w=w, h=h, r=corner_radius, fill='none')
    return g


  def plot(self, pos:Vec=(0,0), size:Vec=(512,1024), *,
   x:'PlotAxis'=None, y:'PlotAxis'=None,
   series:Sequence['PlotSeries']=(),
   title:str=None,
   title_h:Num=14,
   axis_label_h:Num=12,
   tick_h:Num=10,
   tick_len:Num=4,
   legend_pad:Num=8,
   legend_w:Num=128,
   legend_h:Num=16,
   corner_radius:VecOrNum=None,
   symmetric_xy=False,
   dbg=False,
   **attrs:Any) -> 'Plot':

    return self.child(Plot,
      pos=pos, size=size,
      x=x, y=y,
      series=series,
      title=title,
      title_h=title_h,
      axis_label_h=axis_label_h,
      tick_h=tick_h,
      tick_len=tick_len,
      legend_pad=legend_pad,
      legend_w=legend_w,
      legend_h=legend_h,
      corner_radius=corner_radius,
      symmetric_xy=symmetric_xy,
      dbg=dbg,
      **attrs)



class Svg(SvgBranch):
  '''
  Svg is an XmlWriter class that outputs SVG code.
  '''

  tag = 'svg'

  def __init__(self, pos:Vec=None, size:VecOrNum=None, *, x:Dim=None, y:Dim=None, w:Dim=None, h:Dim=None,
   vx:Num=0, vy:Num=0, vw:Num=None, vh:Num=None, _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if size is not None:
      assert w is None
      assert h is None
      if isinstance(size, tuple):
        w, h = size
      else:
        w = h = size
    self.x = x
    self.y = y
    self.w = w
    self.h = h
    self.vx = vx
    self.vy = vy
    self.vw = vw
    self.vh = vh
    self.viewBox = fmt_viewBox(vx, vy, vw, vh)
    kwargs = { # Put the xml nonsense up front.
      'xmlns': 'http://www.w3.org/2000/svg',
      'xmlns:xlink': 'http://www.w3.org/1999/xlink', # Safari does not yet support SVG 2.
      **kwargs
    }
    add_opt_attrs(kwargs, x=x, y=y, width=w, height=h, viewBox=self.viewBox)
    super().__init__(_counters=_counters, attrs=attrs, **kwargs)


# SVG branch elements.

class ClipPath(SvgBranch):
  tag = 'clipPath'

class Defs(SvgBranch):
  tag = 'defs'


class G(SvgBranch):
  tag = 'g'

  def __init__(self, *children:Any, transform:Iterable[str]='',
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    t:str = ''
    if isinstance(transform, str):
      t = transform
    else:
      t = ' '.join(transform)
    if t:
      kwargs['transform'] = t
    super().__init__(*children, _counters=_counters, attrs=attrs, **kwargs)


class Marker(SvgBranch):
  tag = 'marker'

  def __init__(self, *children:Any, id:str='', pos:Vec=None, size:VecOrNum=None, x:Num=None, y:Num=None, w:Num=None, h:Num=None,
   vx:Num=0, vy:Num=0, vw:Num=None, vh:Num=None, markerUnits='strokeWidth', orient:str='auto',
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if not id: raise ValueError(f'Marker requires an `id` string')
    if pos is not None:
      assert x is None
      assert y is None
      x, y = pos
    if size is not None:
      assert w is None
      assert h is None
      w, h = unpack_VecOrNum(size)
    add_opt_attrs(kwargs, id=id, refX=fmt_num(x), refY=fmt_num(y), markerWidth=fmt_num(w), markerHeight=fmt_num(h),
      viewBox=fmt_viewBox(vx, vy, vw, vh), markerUnits=markerUnits, orient=orient)
    super().__init__(*children, _counters=_counters, attrs=attrs, **kwargs)


class Symbol(SvgBranch):
  tag = 'symbol'

  def __init__(self, *children:Any, id:str, vx:Num=None, vy:Num=None, vw:Num=None, vh:Num=None,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    if vx is None: vx = 0
    if vy is None: vy = 0
    # TODO: figure out if no viewBox is legal and at all useful.
    assert vw >= 0 # type: ignore
    assert vh >= 0 # type: ignore
    add_opt_attrs(kwargs, id=id, viewBox=f'{vx} {vy} {vw} {vh}')
    super().__init__(*children, _counters=_counters, attrs=attrs, **kwargs)



# Plots.

Plotter = Callable[[G, PointTransform, Any], None]

def circle_plotter(r:Num=1, **attrs:Any) -> Plotter:
  attrs.setdefault('cl', 'series')
  def plotter(g:G, transform:PointTransform, point:Tuple) -> None:
    g.circle(transform(point), r=r, title=', '.join(str(f) for f in point))
  return plotter



class PlotSeries:

  name:str
  legend:str
  bounds:Optional[BoundsF2] = None # Overridden by subclasses.

  def render(self, plot:'Plot', series:G) -> None: raise NotImplementedError


class BarSeries(PlotSeries):

  def __init__(self, name:str, points:Iterable[Tuple], numeric:bool, legend:str='', width=1.0, plotter:Optional[Plotter]=None,
   title_fmt:Optional[Callable[[Tuple], str]]=None, **attrs:Any) -> None:

    self.name = name
    self.points = list(points)
    self.numeric = numeric
    self.legend = legend or name
    self.width = width
    self.plotter = plotter
    self.title_fmt = title_fmt
    attrs.setdefault('cl', name)
    self.attrs = attrs
    self.bounds:Optional[Tuple[F2, F2]] = None

    def float_from(val:Any, label:str) -> float:
      try: return float(val)
      except TypeError as e: raise TypeError(f'BarSeries received non-numeric point {label}: {val!r}') from e

    if self.points:
      for p in self.points:
        x = p[0]
        y = p[1]
        if numeric:
          x_min = x_max = float_from(x, 'key')
        y_min = y_max = float_from(y, 'value')
        break # Get first value, then iterate again.
      for p in self.points:
        x = p[0]
        y = p[1]
        if numeric:
          x = float_from(x, 'key')
          x_min = min(x_min, x)
          x_max = max(x_max, x)
        y = float_from(y, 'value')
        y_min = min(y_min, y)
        y_max = max(y_max, y)
      if numeric:
        half_w = self.width * 0.5
        self.bounds = ((x_min - half_w, 0.0), (x_max + half_w, y_max))
      else:
        self.bounds = ((0.0, 0.0), (float(len(self.points)), y_max))


  def render(self, plot:'Plot', series:G) -> None:
    assert self.plotter is None # TODO
    y0 = plot.y.transform(0)
    w = plot.x.scale * self.width
    # Do not place the bars in a group because we want to be able to z-index bars across multiple series.
    for i, p in enumerate(self.points):
      if not self.numeric:
        p = (i+0.5, p[1], p[0])
      (x_mid, y) = plot.transform(p)
      x_low = x_mid - w*0.5
      assert p[1] >= 0, f'negative bar values are not yet supported: {p!r}'
      kwargs:Dict[str,Any] = {}
      if self.title_fmt:
        kwargs['title'] = self.title_fmt(p)
      series.rect(x=x_low, y=y, w=w, h=y0-y, z_index=y, **kwargs, **self.attrs) # TODO: Custom format of title?


class XYSeries(PlotSeries):

  def __init__(self, name:str, points:Iterable[Tuple], legend:str='', plotter:Optional[Plotter]=circle_plotter(), **attrs:Any) -> None:
    self.name = name
    self.legend = legend or name
    self.points = list(points)
    self.plotter = plotter
    attrs.setdefault('cl', name)
    self.attrs = attrs
    self.bounds:Optional[Tuple[F2, F2]] = None
    if self.points:
      x, y = self.points[0][:2]
      x_min = x_max = float(x)
      y_min = y_max = float(y)
      for p in self.points:
        x = float(p[0])
        y = float(p[1])
        x_min = min(x_min, x)
        x_max = max(x_max, x)
        y_min = min(y_min, y)
        y_max = max(y_max, y)
      self.bounds = ((x_min, y_min), (x_max, y_max))


  def render(self, plot:'Plot', series:G) -> None:
    # TODO: collect and return out-of-bounds points.
    assert self.plotter is not None
    with series.g(**self.attrs) as g:
      for p in self.points: self.plotter(g, plot.transform, p)



class LineSeries(XYSeries):

  def __init__(self, name:str, points:Iterable[Tuple], legend:str='', plotter:Plotter=None, use_segments=False, **attrs:Any) -> None:
    self.use_segments = use_segments
    super().__init__(name=name, points=points, legend=legend, plotter=plotter, **attrs)


  def render(self, plot:'Plot', series:G) -> None:
    with series.g(**self.attrs) as g:
      if self.use_segments:
        for (a, b) in window_iter(plot.transform(p) for p in self.points):
          g.line(a, b)
      else:
        g.polyline(points=iter_unique(plot.transform(p) for p in self.points), fill='none')
        # TODO: option to fill polylines.
      if self.plotter is not None:
        for p in self.points: self.plotter(g, plot.transform, p)



class PlotAxis:

  def __init__(self, *,
   length:Num=0, # Screen length.
   min:Optional[Num]=None, max:Optional[Num]=None,
   visible_origin=False,
   symmetric=False,
   show_grid=True, grid:Iterable[Num]=(), grid_step:Num=0.0, grid_min:Num=16,
   show_ticks=True, ticks:Iterable[Num]=(), tick_step:Num=0.0, tick_space:Num=1,
   tick_w:Num=16, tick_fmt:Optional[TickFmt]=None) -> None:

    self.length = float(length) # For <=0 the screen length is calculated automatically.
    self._min = None if min is None else float(min)
    self._max = None if max is None else float(max)
    self.visible_origin = visible_origin
    self.symmetric = symmetric
    self.show_grid = show_grid
    self.grid = list(grid)
    self.grid_step = float(grid_step)
    self.grid_min = float(grid_min)
    self.show_ticks = show_ticks
    self.ticks = list(ticks)
    self.tick_step = float(tick_step)
    self.tick_space = float(tick_space)
    self.tick_w = float(tick_w)
    self.tick_fmt = tick_fmt

    # These attributes are filled in by Plot init.
    self.idx = -1
    self.data_len = 0.0
    self.min = 0.0
    self.max = 0.0
    self.partial_tick_len = 0.0
    self.transform = _axis_transform_dummy
    self.scale = 0.0

  def calc_min_max(self, data_bounds:Optional[BoundsF2]) -> None:
    # Determine data data_bounds.
    _min:Optional[float] = self._min
    _max:Optional[float] = self._max
    if _min is None: _min = 0.0   if data_bounds is None else data_bounds[0][self.idx]
    if _max is None: _max = _min if data_bounds is None else data_bounds[1][self.idx]
    if _max <= _min: _max = _min + 1.0
    if self.visible_origin:
      if _min > 0.0:   _min = 0.0
      elif _max < 0.0: _max = 0.0
    if self.symmetric:
      _max = max(_max, -_min)
      _min = -_max
    self.min = _min
    self.max = _max

  def calc_layout(self, plot_size:F2, title_h:float, tick_len:float, tick_h:float) -> None:
    assert self.length > 0, self.length
    self.data_len = self.max - self.min
    self.scale = self.length / self.data_len

    # Calculate tick step.
    tick_min_screen_step = (self.tick_w*1.5) if (self.idx == 0) else (tick_h*2.0)
    tick_mult = 1.0
    frac_w = 0
    fmt_w = 0
    if self.tick_step <= 0 and tick_min_screen_step > 0:
      step1, tick_mult = self.choose_step(tick_min_screen_step)
      self.tick_step = step1 * tick_mult
    if self.tick_step > 0 and not self.tick_fmt:
      exp = floor(log10(self.tick_step))
      frac_w = max(0, -exp)
      fmt_w = max(
        len(f'{self.min:,.{frac_w}f}'),
        len(f'{self.max:,.{frac_w}f}'))

      pad_str = f'{10**fmt_w:,.{frac_w}f}'[-fmt_w:] # Longer than necessary. Take pad chars from right to left.
      def tick_fmt(val:float) -> Any:
        s = f'{val:,.{frac_w}f}'
        return TSpan(pad_str[:-len(s)], cl='zpad'), s

      self.tick_fmt = tick_fmt

    # Calculate grid step.
    if self.grid_step <= 0:
      step1, mult = self.choose_step(min_screen_step=self.grid_min)
      if mult == 2 and tick_mult == 5: # Ticks will misalign to grid; bump grid to 2.5.
        mult = 2.5
      self.grid_step = step1 * mult

    if self.ticks:
      if not self.grid:
        self.grid = self.ticks
    else: # Calculate ticks.
      ti, tr = divmod(self.min, self.tick_step)
      if tr > 0.1: ti += 1 # If the remainder is visually significant, skip the first tick.
      t_start = ti*self.tick_step
      self.ticks = [t for t in NumRange(t_start, self.max, step=self.tick_step, closed=True)]
      if not self.grid: # Calculate grid.
        g_start = (self.min//self.grid_step + 1) * self.grid_step # Skip line index 0 because it is always <= low border.
        self.grid = [g for g in NumRange(g_start, self.max, self.grid_step)]



  def choose_step(self, min_screen_step:float) -> Tuple[float, float]:
    assert self.data_len > 0
    assert min_screen_step > 0
    cram_num = max(1.0, self.length // min_screen_step) # Maximum number of ticks that could be placed.
    assert cram_num > 0, (cram_num, self.length, min_screen_step)
    cram_step = self.data_len / cram_num # Minimum fractional data step.
    exp = floor(log10(cram_step))
    step1 = float(10**exp) # Low estimate of step.
    for mult in (1.0, 2.0, 5.0):
      step = step1 * mult
      if step >= cram_step: return step1, mult
    return step1, 10.0



class Plot(G):

  def __init__(self, *children:Any,
   pos:Vec=(0,0), size:Vec=(512,1024),
   x:PlotAxis=None, y:PlotAxis=None,
   series:Sequence[PlotSeries]=(),
   title:str=None,
   title_h:Num=14,
   axis_label_h:Num=12,
   tick_h:Num=10,
   tick_len:Num=4,
   legend_pad:Num=8,
   legend_w:Num=64,
   legend_h:Num=16,
   corner_radius:VecOrNum=None,
   symmetric_xy=False,
   dbg=False,
   _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs) -> None:

    attrs = attrs or {}
    pos = f2_for_vec(pos)
    # Initialize as `g` element.
    kwargs.setdefault('cl', 'plot')
    super().__init__(*children, transform=translate(*pos), _counters=_counters, attrs=attrs, **kwargs)

    self.pos = pos
    self.size = size = f2_for_vec(size)
    self.x = x = PlotAxis() if x is None else x
    self.y = y = PlotAxis() if y is None else y
    self.series = series
    self.title = title
    self.title_h = title_h = max(
      (0.0 if title is None else float(title_h)),
      (0.0 if (not y.show_ticks) else float(tick_h)))
    self.tick_len = tick_len = float(tick_len)
    self.axis_label_h = axis_label_h = float(axis_label_h)
    self.tick_h = tick_h = float(tick_h)
    self.legend_pad = float(legend_pad)
    self.legend_w = float(legend_w)
    self.legend_h = float(legend_h)
    self.corner_radius = corner_radius
    self.symmetric_xy = symmetric_xy
    self.w = size[0]
    self.h = size[1]

    x.idx = 0
    y.idx = 1
    x.partial_tick_len = tick_h
    y.partial_tick_len = x.tick_w

    data_bounds = reduce(expand_opt_bounds, (s.bounds for s in series), None)

    x.calc_min_max(data_bounds=data_bounds)
    y.calc_min_max(data_bounds=data_bounds)

    if symmetric_xy:
      x.min = min(x.min, y.min)
      y.min = x.min
      x.max = max(x.max, y.max)
      y.max = x.max

    # Calculate per-axis length as necessary.

    boundary_pad = 1 # Otherwise right/bottom can disappear.
    if x.length <= 0:
      y_total_tick_w = (tick_len + y.tick_space + y.tick_w) if y.show_ticks else 0.0
      x.length = size[0] - boundary_pad - max(x.tick_w, y_total_tick_w)
    if y.length <= 0:
      x_total_tick_h = (tick_len + x.tick_space + tick_h) if x.show_ticks else 0.0
      y.length = size[1] - boundary_pad - title_h - x_total_tick_h - (legend_pad+legend_h if legend_h else 0)

    x.calc_layout(plot_size=size, title_h=title_h, tick_len=tick_len, tick_h=tick_h)
    y.calc_layout(plot_size=size, title_h=title_h, tick_len=tick_len, tick_h=tick_h)

    self.data_size = data_size = (x.data_len, y.data_len)
    self.grid_size = grid_size = (x.length, y.length)
    self.scale = (x.scale, y.scale)

    assert x is not None
    assert y is not None
    ax = x # Hack around mypy.
    ay = y

    def transform(point:Sequence) -> F2:
      'Translate a point to appear coincident with the data space.'
      px = float(point[0])
      py = float(point[1])
      return (round(ax.scale*(px-ax.min), 1), round(ay.scale*(ay.data_len - (py-ay.min)), 1))

    def x_transform(val:Num) -> float: return round(ax.scale*(float(val) - ax.min), 1)
    def y_transform(val:Num) -> float: return round(ay.scale*(ay.data_len - (float(val)-ay.min)), 1)


    self.transform = transform
    x.transform = x_transform
    y.transform = y_transform

    if dbg:
      def dbg_rect(pos:Vec, size:Vec, stroke:str=None, fill:str=None, parent=self) -> None:
        parent.rect(pos, size, cl='DBG', stroke=stroke, fill=fill, opacity=0.2)
    else:
      def dbg_rect(pos:Vec, size:Vec, stroke:str=None, fill:str=None, parent=self) -> None: pass

    # Contents.
    self.style(
      _plot_style,
      f'text.title {{ font-size: {title_h}px; }}\n',
      f'text.axis-label {{ font-size: {axis_label_h}px; }}\n',
      f'text.tick {{ font-size: {tick_h}px; }}\n',
      f'tspan.zpad {{ opacity: 0; }}\n',
      f'g.legend text {{ font-size: {legend_h}px; }}\n',
    )

    dbg_rect((0, 0), self.size, stroke='#000')

    # Title.
    dbg_rect((0, 0), (self.w, title_h), fill='#F00')

    if self.title is not None:
      self.text(self.title, pos=(0, 0), cl='title')

    self.area = area = self.g(transform=translate(0, title_h)) # Leave area open.

    # Clip path is is defined to match grid.
    clip_path_id = area.gen_id()
    self.plot_clip_path = f'url(#{clip_path_id})'
    with area.clipPath(id=clip_path_id) as clipPath:
      clipPath.rect(size=grid_size, r=corner_radius)

    # Grid.
    # TODO: if we are really going to support rounded corners then the border rect should clip the interior lines.
    with area.g(cl='grid') as g:
      if x.show_grid:
        g_start_x = (x.min//x.grid_step + 1) * x.grid_step # Skip line index 0 because it is always <= low border.
        for gx in x.grid: # X axis.
          tgx = x.transform(gx)
          g.line((tgx, 0), (tgx, y.length)) # Vertical lines.
      if y.show_grid:
        g_start_y = (y.min//y.grid_step + 1) * y.grid_step # Skip line index 0 because it is always <= low border.
        for gy in y.grid:
          tgy = y.transform(gy)
          g.line((0, tgy), (x.length, tgy)) # Horizontal lines.
      g.rect(cl='grid-border', pos=(0,0), size=grid_size, r=corner_radius, fill='none')

    # Axes.
    if y.min <= 0 and y.max >= 0: # Draw X axis.
      y0 = y.transform(0)
      area.line((0, y0), (x.length, y0), cl='axis', id='x-axis')
    if x.min <= 0 and x.max >= 0: # Draw Y axis.
      x0 = x.transform(0)
      area.line((x0, 0), (x0, y.length), cl='axis', id='y-axis')

    def handle_rendered_tick(val:Any) -> Tuple:
      if isinstance(val, str): return (val,)
      try: return tuple(val)
      except TypeError: return (val,)

    # Ticks.
    if x.show_ticks:
      with area.g(cl='tick-x') as g:
        txi, txr = divmod(x.min, x.tick_step)
        if txr > 0.1: txi += 1 # If the remainder is visually significant, skip the first tick.
        t_start_x = txi*x.tick_step
        for _x in x.ticks:
          tx = x.transform(_x)
          ty = y.length
          tb = ty + tick_len
          tty = tb + x.tick_space
          g.line((tx, ty), (tx, tb), cl='tick')
          assert x.tick_fmt is not None
          dbg_rect((tx, tty), (x.tick_w, tick_h), fill='#008', parent=g)
          g.text(*handle_rendered_tick(x.tick_fmt(_x)), pos=(tx, tty), cl='tick')
    if y.show_ticks:
      with area.g(cl='tick-y') as g:
        tyi, tyr = divmod(y.min, y.tick_step)
        if tyr > 0.1: tyi += 1 # If the remainder is visually significant, skip the first tick.
        t_start_y = tyi*y.tick_step
        for _y in y.ticks:
          tx = x.length
          tr = tx + tick_len
          ttx = tr + y.tick_space
          ty = y.transform(_y)
          g.line((tx, ty), (tr, ty), cl='tick')
          assert y.tick_fmt is not None
          dbg_rect((ttx, ty-tick_h*0.75), (y.tick_w, tick_h), fill='#080', parent=g)
          g.text(*handle_rendered_tick(y.tick_fmt(_y)), pos=(ttx, ty), cl='tick')

    # Legend.
    leg_h = self.legend_h
    if leg_h > 0:
      leg_y = size[1] - title_h - self.legend_h - 1
      with area.g(cl='legend') as legend_g:
        for i, s in enumerate(series):
          with legend_g.g(cl='legend-'+s.name) as g:
            leg_x = 1 + self.legend_w*i
            g.rect(pos=(leg_x, leg_y), size=(leg_h, leg_h), cl=s.name)
            text_x = leg_x + leg_h * 1.25
            text_y = leg_y + leg_h * 0.5
            g.text(s.legend, pos=(text_x, text_y))

    # Series.
    with area.g(cl='series', clip_path=self.plot_clip_path) as series_g:
      for s in series:
        s.render(self, series=series_g)


_plot_style = '''
text {
  stroke: none;
  fill: currentColor;
}
text.title {
  text-anchor: start;
  alignment-baseline: hanging;
}
g.tick-x text.tick {
  white-space: pre;
  text-anchor: start;
  alignment-baseline: hanging;
}
g.tick-y text.tick {
  white-space: pre;
  text-anchor: start;
  alignment-baseline: alphabetic;
}
g.legend text {
  text-anchor: start;
  alignment-baseline: central;
}
'''


# Elements.

class SvgTitle(SvgWriter):
  tag = 'title'


# Transforms.

def scale(x:Num=1, y:Num=1) -> str: return f'scale({x},{y})'
def rotate(degrees:Num, x:Num=0, y:Num=0) -> str: return f'rotate({fmt_num(degrees)},{fmt_num(x)},{fmt_num(y)})'
def translate(x:Num=0, y:Num=0) -> str: return f'translate({fmt_num(x)},{fmt_num(y)})'


# Miscellaneous.


_path_command_lens = {
  'A' : 7,
  'M' : 2,
  'm' : 2,
  'L' : 2,
  'l' : 2,
  'Z' : 0,
  'z' : 0
}


valid_units = frozenset({'em', 'ex', 'px', 'pt', 'pc', 'cm', 'mm', '%', ''})

def _validate_unit(unit: str):
  'Ensure that `unit` is a valid unit string.'
  if unit not in valid_units:
    raise Exception(f'Invalid SVG unit: {unit!r}; should be one of {sorted(valid_units)}')


def fmt_viewBox(vx:Optional[Num], vy:Optional[Num], vw:Optional[Num], vh:Optional[Num]) -> Optional[str]:
  if vw is None and vh is None:
    return None
  else:
    if vx is None: vx = 0
    if vy is None: vy = 0
    assert vw is not None and vw > 0
    assert vh is not None and vh > 0
    return f'{fmt_num(vx)} {fmt_num(vy)} {fmt_num(vw)} {fmt_num(vh)}'


@overload
def fmt_num(n:Num) -> str: ...
@overload
def fmt_num(n:None) -> None: ...

def fmt_num(n:Optional[Num]) -> Optional[str]:
  'Remove trailing ".0" from floats that can be represented as integers.'
  if n is None: return None
  if isinstance(n, float):
    i = int(n)
    if i == n: return str(i)
  return str(n)


def f2_for_vec(v:Vec) -> F2:
  x, y = v
  return (float(x), float(y))


def unpack_VecOrNum(vn:VecOrNum) -> Tuple[float, float]:
  if isinstance(vn, tuple):
    x, y = vn
    return float(x), float(y)
  else:
    s = float(vn)
    return (s, s)


def expand_opt_bounds(l:Optional[BoundsF2], r:Optional[BoundsF2]) -> Optional[BoundsF2]:
  if l is None: return r
  if r is None: return l
  (llx, lly), (lhx, lhy) = l
  (rlx, rly), (rhx, rhy) = r
  return ((min(llx, rlx), min(lly, rly)), (max(lhx, rhx), max(lhy, rhy)))


alignment_baselines = {
  'after-edge',
  'alphabetic',
  'auto',
  'baseline',
  'before-edge',
  'central',
  'hanging',
  'ideographic',
  'inherit',
  'mathematical',
  'middle',
  'text-after-edge',
  'text-before-edge',
}



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Various exception classes.
'''


class ConflictingValues(ValueError):
  '''
  Raised when an incoming value collides with an existing value.
  In one sense, it is both a KeyError and a ValueError.
  '''

class DeleteNode(Exception):
  'Signals a traverser to delete the current node.'

class FlattenNode(Exception):
  'Signals a traverser to delete the current node.'

class OmitNode(Exception):
  'Signals a transformer to omit the current node.'


class MultipleMatchesError(KeyError):
  'Raised when a query matches multiple children.'


class NoMatchError(KeyError):
  'Raised when a query matches no children.'




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Callable, Hashable, Iterable, Set, TypeVar


_H = TypeVar('_H', bound=Hashable)

def visit_nodes(start_nodes:Iterable[_H], visitor:Callable[[_H], Iterable[_H]]) -> Set[_H]:
  '''
  Starting with `start_nodes`, call `visitor` with each node.
  `visitor` should return discovered nodes to be visited.
  Each node is visited exactly once.
  The set of all visited nodes is returned.
  '''
  remaining = set(start_nodes)
  visited = set()
  while remaining:
    node = remaining.pop()
    visited.add(node)
    discovered = visitor(node)
    remaining.update(n for n in discovered if n not in visited)
  return visited



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

# count file extensions found in a directory tree.

# TODO:
# implement path arguments
# multiple arguments should cause each directory to be recursed
# stats should be printed for each directory when completed, then total stats when all done.

# implement ext listing (-l)
#  usage example: count-ext -l "mp3 m4a"
#  this would cause the relative path of each file matching the given extensions to be listed
#  in this case don't print stats, so that the results can be piped to other utilities


from argparse import ArgumentParser
from typing import DefaultDict

from ..fs import path_ext, walk_files


def main() -> None:
  parser = ArgumentParser(description='Count lines of source code.')
  parser.add_argument('paths', nargs='+', help='Directories to explore.')
  args = parser.parse_args()

  ext_counts = DefaultDict[str,int](int)

  for path in walk_files(*args.paths):
    ext = path_ext(path)
    ext_counts[ext] += 1

  empty_name = '[none]'
  max_len = max(len(empty_name), max(len(k) for k in ext_counts))

  for k in sorted(ext_counts.keys()):
    v = ext_counts[k]
    print('{0:{1}}: {2}'.format(k or empty_name, max_len, v))


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Counter, Dict, List, NamedTuple, Tuple


class UnicodeCategory(NamedTuple):
  key:str
  name:str
  desc:str
  subcategories:Tuple[str, ...]


def _mk_cat(key:str, name:str, desc:str) -> UnicodeCategory:
  subs = desc.split(' | ')
  return UnicodeCategory(key=key, name=name, desc=desc, subcategories=tuple(subs if len(subs) > 1 else []))


unicode_categories:List[UnicodeCategory] = [ # taken directly from: http://www.unicode.org/reports/tr44/#General_Category_Values.
  _mk_cat('Lu', 'Uppercase_Letter',      'An uppercase letter'),
  _mk_cat('Ll', 'Lowercase_Letter',      'A lowercase letter'),
  _mk_cat('Lt', 'Titlecase_Letter',      'A digraphic character, with first part uppercase'),
  _mk_cat('LC', 'Cased_Letter',          'Lu | Ll | Lt'),
  _mk_cat('Lm', 'Modifier_Letter',       'A modifier letter'),
  _mk_cat('Lo', 'Other_Letter',          'other letters, including syllables and ideographs'),
  _mk_cat('L',  'Letter',                'Lu | Ll | Lt | Lm | Lo'),
  _mk_cat('Mn', 'Nonspacing_Mark',       'A nonspacing combining mark (zero advance width)'),
  _mk_cat('Mc', 'Spacing_Mark',          'A spacing combining mark (positive advance width)'),
  _mk_cat('Me', 'Enclosing_Mark',        'An enclosing combining mark'),
  _mk_cat('M',  'Mark',                  'Mn | Mc | Me'),
  _mk_cat('Nd', 'Decimal_Number',        'A decimal digit'),
  _mk_cat('Nl', 'Letter_Number',         'A letterlike numeric character'),
  _mk_cat('No', 'Other_Number',          'A numeric character of other type'),
  _mk_cat('N',  'Number',                'Nd | Nl | No'),
  _mk_cat('Pc', 'Connector_Punctuation', 'A connecting punctuation mark, like a tie'),
  _mk_cat('Pd', 'Dash_Punctuation',      'A dash or hyphen punctuation mark'),
  _mk_cat('Ps', 'Open_Punctuation',      'An opening punctuation mark (of a pair)'),
  _mk_cat('Pe', 'Close_Punctuation',     'A closing punctuation mark (of a pair)'),
  _mk_cat('Pi', 'Initial_Punctuation',   'An initial quotation mark'),
  _mk_cat('Pf', 'Final_Punctuation',     'A final quotation mark'),
  _mk_cat('Po', 'Other_Punctuation',     'A punctuation mark of other type'),
  _mk_cat('P',  'Punctuation',           'Pc | Pd | Ps | Pe | Pi | Pf | Po'),
  _mk_cat('Sm', 'Math_Symbol',           'A symbol of mathematical use'),
  _mk_cat('Sc', 'Currency_Symbol',       'A currency sign'),
  _mk_cat('Sk', 'Modifier_Symbol',       'A non-letterlike modifier symbol'),
  _mk_cat('So', 'Other_Symbol',          'A symbol of other type'),
  _mk_cat('S',  'Symbol',                'Sm | Sc | Sk | So'),
  _mk_cat('Zs', 'Space_Separator',       'A space character (of various non-zero widths)'),
  _mk_cat('Zl', 'Line_Separator',        'U+2028 LINE SEPARATOR only'),
  _mk_cat('Zp', 'Paragraph_Separator',   'U+2029 PARAGRAPH SEPARATOR only'),
  _mk_cat('Z',  'Separator',             'Zs | Zl | Zp'),
  _mk_cat('Cc', 'Control',               'A C0 or C1 control code'),
  _mk_cat('Cf', 'Format',                'A format control character'),
  _mk_cat('Cs', 'Surrogate',             'A surrogate code point'),
  _mk_cat('Co', 'Private_Use',           'A private-use character'),
  _mk_cat('Cn', 'Unassigned',            'A reserved unassigned code point or a noncharacter'),
  _mk_cat('C',  'Other',                 'Cc | Cf | Cs | Co | Cn'),
  _mk_cat('CE', 'Other_Encodable',       'Cc | Cf | Co | Cn'), # This category is a Legs invention; excludes unencodable surrogates.
]


unicode_category_aliases:Dict[str,UnicodeCategory] = { cat.key : cat for cat in unicode_categories }

def _add_aliases() -> None:
  unicode_category_aliases.update((cat.name, cat) for cat in unicode_categories)
  # add first-word aliases wherever they are unambiguous.
  first_word_counts:Counter[str] = Counter()
  first_word_categories = {}
  for cat in unicode_categories:
    first = cat.name.partition('_')[0]
    if first in unicode_category_aliases: continue
    first_word_counts[first] += 1
    first_word_categories[first] = cat
  unicode_category_aliases.update(p for p in first_word_categories.items() if first_word_counts[p[0]] == 1)

_add_aliases()





from typing import List, Sequence, Tuple, TypeVar


_T = TypeVar('_T')


def ranges_without_common_ends(seq_a:Sequence[_T], seq_b:Sequence[_T]) -> Tuple[range, range]:
  pre_idx = 0
  for pre_idx, (el_a, el_b) in enumerate(zip(seq_a, seq_b)):
    if el_a != el_b: break
  post_a = len(seq_a)
  post_b = len(seq_b)
  for idx_a, idx_b in zip(reversed(range(pre_idx, len(seq_a))), reversed(range(pre_idx, len(seq_b)))):
    if seq_a[idx_a] != seq_b[idx_b]:
      post_a = idx_a + 1
      post_b = idx_b + 1
      break
  return (range(pre_idx, post_a), range(pre_idx, post_b))


def trim_common_ends(a:List[_T], b:List[_T]) -> Tuple[List[_T], List[_T]]:
  r_a, r_b = ranges_without_common_ends(a, b)
  return a[r_a.start:r_a.stop], b[r_b.start:r_b.stop]


#def adjust_boundaries(seq:Sequence[_T],


# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from time import time as now
from typing import Callable, Optional

from ..fs import file_size, path_exists, remove_path
from ..io import confirm, outL, outZ, stderr, stdin
from ..string import format_byte_count
from ..task import UnexpectedExit, run


PackFn = Callable[[str, str], None]


def main() -> None:
  parser = ArgumentParser(description='Compress files.')
  parser.add_argument('-overwrite', action='store_true', help='remove any existing archive file before writing.')
  parser.add_argument('-quiet', action='store_true', help='do not print compression statistics.')
  parser.add_argument('-level', help='compression level (format specific).')
  parser.add_argument('-br',  action='store_true', help=f'compress using brotli.')
  parser.add_argument('-gz',  action='store_true', help=f'compress using gzip.')
  parser.add_argument('-xz',  action='store_true', help=f'compress using xz.')
  parser.add_argument('-zst', action='store_true', help=f'compress using zstd.')
  parser.add_argument('paths', nargs='+', help='Paths to pack.')
  args = parser.parse_args()

  if not any((args.br, args.gz, args.xz, args.zst)):
    exit(f'error: please specify at least one compression format: {format_flags}.')

  show_stats = not args.quiet
  kwargs = dict(overwrite=args.overwrite, level=args.level, show_stats=show_stats)
  for path in args.paths:
    if args.br:   pack(path, ext='.br', **kwargs)
    if args.gz:   pack(path, ext='.gz', **kwargs)
    if args.xz:   pack(path, ext='.xz', **kwargs)
    if args.zst:  pack(path, ext='.zst', **kwargs)

def pack(path:str, ext:str, overwrite:bool, level:Optional[str], show_stats:bool) -> None:
  dst = path + ext
  pack_fn = formats[ext]
  if path_exists(dst, follow=False):
    if not overwrite:
      if stdin.isatty():
        if not confirm(f'archive path {dst!r} exists; remove it?'): exit(1)
      else:
        exit(f'archive path exists: {dst!r}')
    remove_path(dst)
  if show_stats: outZ(dst, ':')
  try:
    start = now()
    pack_fn(path, level=level)
    duration = now() - start
  except Exception as e:
    if show_stats: outL()
    if isinstance(e, UnexpectedExit): exit(1)
    else: raise
  else:
    orig_size = file_size(path)
    pack_size = file_size(dst)
    ratio = (pack_size / orig_size) if orig_size else 0.0
    if show_stats: outL(f' {format_byte_count(orig_size)} -> {format_byte_count(pack_size)} ({ratio:.0%}; {duration:0.2f}s)')


def br(path:str, level:str=None) -> None:
  if not level: level = '5'
  run(['brotli', '--keep', f'-{level}', path])

def gz(path:str, level:str=None) -> None:
  if not level: level = '5'
  run(['gzip', '--keep', f'-{level}', path])

def xz(path:str, level:str=None) -> None:
  if not level: level = '4'
  run(['xz', '--keep', f'-{level}', '--threads=0', path])

def zst(path:str, level:str=None) -> None:
  if not level: level = '6'
  run(['zstd', '--keep', f'-{level}', '--threads=0', '-q', path], err=stderr)


formats = {
  '.br'  : br,
  '.gz'  : gz,
  '.xz'  : xz,
  '.zst' : zst,
}

format_flags = ', '.join(ext.replace('.', '-') for ext in formats)



#!/usr/bin/env python3
# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import os
import re
from argparse import ArgumentParser
from cProfile import Profile
from os.path import dirname
from pstats import Stats
from sys import argv, exc_info, path as sys_path, stderr, stdout
from typing import Any, Dict, Iterable, List, Optional, Set, TextIO, Tuple, Union

from pithy.fs import path_for_cmd
from pithy.path import path_rel_to_current_or_abs


def main() -> None:
  sort_keys_l = ', '.join(sort_keys)
  sort_keys_desc = f'sort keys: {sort_keys_l}.'
  parser = ArgumentParser(description='Run a python script under the Python cProfile profiler.')
  parser.add_argument('-sort', nargs='+', default=['cumulative', 'filename', 'name'], help=sort_keys_desc)
  parser.add_argument('-filter', nargs='+', default=[], help='filtering clauses.')
  parser.add_argument('-output', default='<stderr>', help='Output file; defaults to <stderr>.')
  parser.add_argument('cmd', nargs='+', help='the command to run.')
  args = parser.parse_args()

  cmd = args.cmd
  cmd_path = path_for_cmd(cmd[0])

  output:TextIO
  if args.output == '<stderr>': output = stderr
  elif args.output in ('-', '<stdout>'): output = stdout
  else: output = open(args.output, 'w')

  def filter_clause(word:Any) -> Any:
    for T in (int, float):
      try: return T(word)
      except ValueError: pass
    return word

  filter_clauses = [filter_clause(word) for word in args.filter]

  with open(cmd_path, 'rb') as f:
    code = compile(f.read(), cmd_path, 'exec')

  globals_ = {
    '__file__': cmd_path,
    '__name__': '__main__',
    '__package__': None,
    '__cached__': None,
  }

  argv[:] = cmd
  # also need to fix the search path to imitate the regular interpreter.
  sys_path[0] = dirname(cmd_path) # not sure if this is right in all cases.

  profile = Profile()
  exit_code = 0
  try:
    profile.runctx(code, globals=globals_, locals=globals_)
  except SystemExit as e:
    exit_code = e.code
  except BaseException:
    from traceback import TracebackException
    exit_code = 1 # exit code that Python returns when an exception raises to toplevel.
    # Format the traceback as it would appear when run standalone.
    traceback = TracebackException(*exc_info())
    # note: the traceback will contain stack frames from the host.
    # this can be avoided with a fixup function, but does not seem necessary at this point. See coven.py for an example.
    #fixup_traceback(traceback)
    print(*traceback.format(), sep='', end='', file=stderr)

  stats = CustomStats(profile, stream=output)
  stats.sort_stats(*args.sort)

  stats.print('\n', '=' * 64, sep='')
  if stats.fcn_list:
    stats.print(f'Ordered by: {stats.sort_type}.\n')
  else:
    stats.print('Random listing order was used.\n')

  stats.display_stats(*filter_clauses)
  #stats.print('\nCallers:')
  #stats.display_callers(*filter_clauses)
  #stats.print('\nCallees:')
  #stats.display_callees(*filter_clauses)

  exit(exit_code)


sort_keys = {
  'calls'       : 'call count',
  'cumtime'     : 'cumulative time',
  'cumulative'  : 'cumulative time',
  'file'        : 'file name',
  'filename'    : 'file name',
  'line'        : 'line number',
  'module'      : 'file name',
  'name'        : 'function name',
  'ncalls'      : 'call count',
  'nfl'         : 'name/file/line',
  'pcalls'      : 'primitive call count',
  'stdname'     : 'standard name',
  'time'        : 'internal time',
  'tottime'     : 'internal time',
}


Func = Tuple[str,int,str]
Selector = Union[str,float,int]


class CustomStats(Stats):

  all_callees:Optional[Dict]
  files:List[TextIO]
  max_name_len:int
  fcn_list:List[Func]
  prim_calls:int
  sort_type:str
  stats:Dict[Func,Any]
  stream:TextIO
  top_level:Set[Func]
  total_calls:int
  total_tt:float


  def print(self, *items:Any, sep=' ', end='\n') -> None:
    print(*items, file=self.stream, sep=sep, end=end)


  def display_stats(self, *amount:Selector) -> None:
    if self.files:
      for filename in self.files:
        self.print(filename)
      self.print()

    for func in self.top_level:
      self.print(get_func_name(func))

    self.print(f'{self.total_calls} function calls ({self.prim_calls} primitive calls) in {self.total_tt:.3f} seconds')
    self.print()
    width, stat_list = self.get_display_list(amount)
    if stat_list:
      self.print_title()
      for func in stat_list:
        self.display_line(func)
      self.print_title()
      self.print()


  def display_callees(self, *amount:Selector) -> None:
    width, stat_list = self.get_display_list(amount)
    if not stat_list: return
    self.calc_callees()
    assert self.all_callees is not None
    self.print_call_heading(width, "CALLEE")
    for func in stat_list:
      if func in self.all_callees:
          self.display_call_line(width, func, self.all_callees[func])
      else:
          self.display_call_line(width, func, {})
    self.print('\n')


  def display_callers(self, *amount:Selector) -> None:
    width, stat_list = self.get_display_list(amount)
    if not stat_list: return
    self.print_call_heading(width, "CALLER")
    for func in stat_list:
      cc, nc, tt, ct, callers = self.stats[func]
      self.display_call_line(width, func, callers, "<-")
    self.print('\n')


  def eval_display_amount(self, sel:Selector, stat_list:List[Func], msg:str) -> Tuple[List[Func], str]:
    new_list = stat_list
    if isinstance(sel, str):
      try:
        rex = re.compile(sel)
      except re.error:
        msg += "   <Invalid regular expression %r>\n" % sel
        return new_list, msg
      new_list = []
      for func in stat_list:
        if rex.search(fmt_func(func)):
            new_list.append(func)
    else:
      count = len(stat_list)
      if isinstance(sel, float) and 0.0 <= sel < 1.0: # type: ignore # Spurious 'Left operand of 'and' is always true'.
        count = int(count * sel + .5)
        new_list = stat_list[:count]
      elif isinstance(sel, int) and 0 <= sel < count:
        count = sel
        new_list = stat_list[:count]
    if len(stat_list) != len(new_list):
      msg += "   List reduced from %r to %r due to restriction <%r>\n" % (
        len(stat_list), len(new_list), sel)
    return new_list, msg


  def get_display_list(self, sel_list:Iterable[Selector]) -> Tuple[int, List[Func]]:
    stat_list = list(self.fcn_list[:]) if self.fcn_list else list(self.stats.keys())

    width = self.max_name_len

    msg = ''
    for selection in sel_list:
      stat_list, msg = self.eval_display_amount(selection, stat_list, msg)

    count = len(stat_list)

    if not stat_list:
      return 0, stat_list
    if count < len(self.stats):
      width = 0
      for func in stat_list:
        if  len(fmt_func(func)) > width:
            width = len(fmt_func(func))
    return width+2, stat_list


  def print_call_heading(self, name_size:int, column_title:str) -> None:
    self.print('Function'.ljust(name_size), column_title)
    # print sub-header only if we have new-style callers
    subheader = False
    for cc, nc, tt, ct, callers in self.stats.values():
      if callers:
        value = next(iter(callers.values()))
        subheader = isinstance(value, tuple)
        break
    if subheader:
      self.print(" "*name_size + "  ncalls  tottime  cumtime")


  def display_call_line(self, name_size:int, source:Func, call_dict:Dict[Func,Any], arrow:str='->') -> None:
    self.print(fmt_func(source).ljust(name_size) + arrow, end=' ')
    if not call_dict:
      self.print()
      return
    clist = sorted(call_dict.keys())
    indent = ""
    for func in clist:
      name = fmt_func(func)
      value = call_dict[func]
      if isinstance(value, tuple):
        nc, cc, tt, ct = value
        if nc != cc:
            substats = '%d/%d' % (nc, cc)
        else:
            substats = '%d' % (nc,)
        substats = '%s %s %s  %s' % (substats.rjust(7+2*len(indent)), f8(tt), f8(ct), name)
        left_width = name_size + 1
      else:
        substats = '%s(%r) %s' % (name, value, f8(self.stats[func][3]))
        left_width = name_size + 3
      self.print(indent*left_width + substats)
      indent = " "


  def print_title(self) -> None:
    self.print('         ncalls  tottime  percall  cumtime  percall filename:lineno(function)')


  def display_line(self, func:Func) -> None:  # hack: should print percentages
    cc, nc, tt, ct, callers = self.stats[func]
    ncalls = str(nc) if nc == cc else f'{nc}/{cc}'
    stt = f8(tt)
    sct = f8(ct)
    ttpc = f8(tt/nc) if nc else ' '*8
    ctpc = f8(ct/nc) if nc else ' '*8
    if all(s == '   0.000' for s in (stt, sct, ttpc, ctpc)): return
    self.print(f'{ncalls:>15} {stt} {ttpc} {sct} {ctpc} {fmt_func(func)}')


def f8(x:float) -> str: return f'{x:8.3f}'


def get_func_name(func:Func) -> str:
  _, _, name = func
  return name


def fmt_func(func:Func) -> str:
  file, line, name = func
  if file == '~' and line == 0: # Special case for built-in functions.
    return name
  if file.startswith(std_prefix):
    file = '|' + file[len(std_prefix):]
  else:
    file = path_rel_to_current_or_abs(file)
  return f'{file}:{line}:{name}'

std_prefix = os.__file__[:-len('os.py')]


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Optional, TypeVar


_T = TypeVar('_T')

def unwrap(optional:Optional[_T]) -> _T:
  if optional is None: raise ValueError('unexpected None')
  return optional



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Produce a schema from input json.
'''

import json
from argparse import ArgumentParser
from sys import stderr, stdin, stdout

from ..fs import walk_files
from ..schema import compile_schema, write_schema


def main() -> None:
  parser = ArgumentParser(description='Count lines of source code.')
  parser.add_argument('-count-atoms', action='store_true')
  parser.add_argument('paths', nargs='*', default=['-'], help='Directories to explore.')
  args = parser.parse_args()

  schema = None
  for path in walk_files(*args.paths, file_exts=['.json']):
    with (stdin if path == '-' else open(path)) as f:
      try: j = json.load(f)
      except json.JSONDecodeError as e:
        print('error: {}: {}'.format(path, e), file=stderr)
        continue
      schema = compile_schema(j, schema=schema)

  write_schema(stdout, schema, count_atoms=args.count_atoms)


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from enum import Enum

from typing import Iterable, Mapping, Union


vertical_bars = ' ' + ''.join(chr(i) for i in range(0x2581, 0x2589))

horizontal_bars = (
  ' '
  '\u258f' # Left one eighth block.
  '\u258e' # Left one quarter block.
  '\u258d' # Left three eighths block.
  '\u258c' # Left half block.
  '\u258b' # Left five eighths block.
  '\u258a' # Left three quarters block.
  '\u2589' # Left seven eighths eighths block.
  '\u2588' # Full block.
)

full_block = '\u2588'

_min = min
_max = max

_Num = Union[int,float]


def chart_seq_inline(values:Iterable[_Num], max:_Num=0) -> str:
  values = tuple(values)
  if not values: return ''
  if max <= 0:
    max = _max(values)
    if max <= 0: return '\u2592' * len(values) # Medium shade block.
  return ''.join(vertical_bars[int(0.5 + (8 * _max(0, _min(1, v/max))))] for v in values)


class ChartMode(Enum):
  Normalized, Total, Cumulative, Ratio = range(4)


def chart_items(m:Mapping, mode=ChartMode.Normalized, threshold=0, sort_by_val=False, width=64) -> str:
  '''
  create a chart from a map, where values are either integers or pairs of integers
  (for ChartModeRatio).
  threshold is a minimum denominator count for ratios, and a minimum ratio otherwise.
  '''

   # rows are of form (sortKey, name, val, ratio). key can be of any type; name and val must be strings.
  rows = []

  if m and mode in (ChartMode.Total, ChartMode.Cumulative):
    total = sum(m.values())
    if mode is ChartMode.Cumulative:
      cum = 0

  elif m and mode is ChartMode.Normalized:
    max_val = max(m.values())
    if max_val <= 0:
      max_val = 1 # hack to prevent divide by zero.

  for k, v in sorted(m.items()):

    if mode is ChartMode.Normalized:
      r = v / max_val
      if r < threshold:
        continue
      val = '{:,}'.format(v)

    elif mode is ChartMode.Total:
      r = v / max(total, 1)
      if r < threshold:
        continue
      val = '{:,}'.format(v)

    elif mode is ChartMode.Cumulative:
      cum += v
      r = cum / total
      if r > 1 - threshold:
        continue
      val = '{:,}'.format(cum)

    elif mode is ChartMode.Ratio:
      if v[0] == 0 or v[1] < threshold:
        continue
      r = v[0] / v[1]
      val = '{:,}/{:,}'.format(*v)


    sort_key = r if sort_by_val else k
    row = (sort_key, str(k), val, r)
    rows.append(row)

  if not rows:
    return ''

  rows.sort(reverse=sort_by_val)

  name_width = max(len(r[1]) for r in rows)
  val_width  = max(len(r[2]) for r in rows)

  lines = [chart_line(n, v, r, name_width=name_width, val_width=val_width, bar_width=width, suffix='\n') for sk, n, v, r in rows]

  return ''.join(lines)


def chart_line(name:str, val:str, ratio:float, name_width:int, val_width:int, bar_width:int, suffix='') -> str:
  'create a string for a single line of a chart.'
  n = f'{name:<{name_width}}'
  v = f'{val:>{val_width}}'
  b = bar_str(ratio, bar_width, pad_right=bool(suffix))

  return '  {} : {}  {:.3f} {}{}'.format(n, v, ratio, b, suffix)


def bar_str(ratio:float, width:int, pad_right=False) -> str:
  'create a string of block characters for the given ratio and width.'
  if ratio > 1:
    return '*' * width

  index = int(ratio * width * 8) # quantize the ratio
  solid_count = index // 8 # number of filled blocks
  fraction_index = index % 8
  solid = full_block * solid_count # string of solid blocks
  fraction = horizontal_bars[fraction_index] if fraction_index else '' # fraction char string
  pad = (' ' * (width - (solid_count + len(fraction)))) if pad_right else ''

  return f'{solid}{fraction}{pad}'



if __name__ == '__main__':
  for mode in (ChartMode.Normalized, ChartMode.Total, ChartMode.Cumulative):
    print(mode)
    m = { i : i for i in range(32) }
    print(chart_items(m, mode=mode))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import os
from datetime import datetime as DateTime, timezone as TimeZone
from gzip import compress as gz_compress, decompress as gz_expand
from io import BytesIO
from mimetypes import guess_type as guess_mime_type
from typing import IO, Any, Callable, Dict, Optional

from boto3 import Session  # type: ignore
from botocore.exceptions import ClientError  # type: ignore

from ..fs import file_status, make_dirs, path_dir, path_ext, path_join, walk_files
from ..json import parse_json, render_json


br_MODE_TEXT = -1
try:
  from brotli import compress as br_compress, decompress as br_expand, MODE_TEXT as br_MODE_TEXT # type: ignore
except ImportError:
  # Brotli compression is supported by major browsers and AWS. Pithy treats it as optional.
  def br_compress(data:bytes, mode:int=0) -> bytes: raise Exception('brotli module failed to import')
  def br_expand(data:bytes) -> bytes: raise Exception('brotli module failed to import')



class S3MockError(Exception): pass


def get_bytes(client:Any, bucket:str, key:str) -> bytes:
  r = client.get_object(Bucket=bucket, Key=key)
  body: bytes = r['Body'].read()
  encoding = r['ResponseMetadata']['HTTPHeaders'].get('content-encoding')
  if encoding is None:
    ext = path_ext(key)
    if ext == '.br': encoding = 'br'
    elif ext == '.gz': encoding = 'gzip'
    else: encoding = 'identity'
  if encoding == 'identity': return body
  elif encoding == 'br': return br_expand(body) # type: ignore
  elif encoding == 'gzip': return gz_expand(body)
  raise ValueError(encoding)


def get_json(client:Any, bucket:str, key:str) -> Any:
  return parse_json(get_bytes(client=client, bucket=bucket, key=key))


def get_text(client:Any, bucket:str, key:str) -> Any:
  return get_bytes(client=client, bucket=bucket, key=key).decode()


def put_bytes(client:Any, data:bytes, bucket:str, key:str, content_encode:str=None, is_utf8_hint=False) -> None:
  '''
  `content_encode` specifies an optional compression encoding, either `gzip` or `br`.
  '''
  content_type, content_encoding = guess_mime_type(key)

  # Validate that the implied encoding appears to be the actual encoding.
  if content_type is None:
    content_type = 'application/octet-stream' # default to binary.
  if content_encoding == 'gzip':
    if data[:2] != b'\x1F\x8B':
      raise Exception(f"save_bytes: filename implies content-type 'gzip' but data does not: {key!r}")
  elif content_encoding == 'br':
    pass # Brotli defines no magic bytes/header.
  elif content_encoding is not None:
    raise Exception(f'unknown content-encoding: {content_encoding!r}')

  # Compress as requested.
  if content_encode is not None:
    if content_encoding is not None:
      raise Exception(f'save_bytes: key {key!r} implies content-type {content_type!r}, but `content_encode` is also specified: {content_encode!r}')
    kwargs:Dict[str,Any] = {}
    if content_encode == 'br' and is_utf8_hint:
      kwargs['mode'] = br_MODE_TEXT
    compress_fn = compressors[content_encode]
    data = compress_fn(data, **kwargs)
    content_encoding = content_encode

  _ = client.put_object(
    Bucket=bucket,
    Key=key,
    ContentType=content_type,
    ContentEncoding=content_encoding or 'identity',
    Body=data)


compressors:Dict[str, Callable[..., Any]] = {
  'br': br_compress,
  'gzip' : gz_compress,
}


def put_json(client:Any, obj:Any, bucket:str, key:str, content_encode:str=None, **kwargs:Any) -> None:
  data = render_json(obj, **kwargs).encode()
  if key.endswith('.br'):
    if content_encode not in (None, 'br'):
      raise Exception(f'put_json: key {key!r} implies `br` compression but content_encode is also specified: {content_encode!r}')
    data = br_compress(data, mode=br_MODE_TEXT)
  elif key.endswith('.gz'):
    data = gz_compress(data)
    if content_encode not in (None, 'gzip'):
      raise Exception(f'put_json: key {key!r} implies `gzip` compression but content_encode is also specified: {content_encode!r}')
  put_bytes(client=client, data=data, bucket=bucket, key=key, content_encode=content_encode)


def put_text(client:Any, text:str, bucket:str, key:str, content_encode:str=None) -> None:
  data = text.encode()
  if key.endswith('.br'):
    if content_encode not in (None, 'br'):
      raise Exception(f'put_text: key {key!r} implies `br` compression but content_encode is also specified: {content_encode!r}')
    data = br_compress(data, mode=br_MODE_TEXT)
  elif key.endswith('.gz'):
    data = gz_compress(data)
    if content_encode not in (None, 'gzip'):
      raise Exception(f'put_text: key {key!r} implies `gzip` compression but content_encode is also specified: {content_encode!r}')
  put_bytes(client=client, data=data, bucket=bucket, key=key, content_encode=content_encode)


class S3Client:
  '''
  Because boto3 generates clients dynamically,
  for now we create a fake base class for mypy to refer to.
  In the future this could become a functioning, statically typed wrapper of the real S3 client.
  '''

  def __init__(self, bucket_paths:Dict[str,str],
    aws_access_key_id:str=None,
    aws_secret_access_key:str=None,
    aws_session_token:str=None,
    config:Any=None,
    region_name:str=None,
   ) -> None:
    raise NotImplementedError

  def get_object(self, Bucket:str, Key:str) -> Dict[str,Any]:
    raise NotImplementedError

  def list_objects_v2(self, Bucket:str) -> Dict[str,Any]:
    raise NotImplementedError

  def put_object(self, Bucket:str, Key:str, ContentType:str, ContentEncoding:str, Body:bytes) -> Dict[str,Any]:
    raise NotImplementedError

  def upload_file(self, path:str, Bucket:str, Key:str) -> Dict[str,Any]:
    raise NotImplementedError



class S3MockClient(S3Client):
  '''
  A local mock of S3 that reads and writes to a provided directory for each bucket.
  '''

  def __init__(self, bucket_paths:Dict[str,str],
    aws_access_key_id:str=None,
    aws_secret_access_key:str=None,
    aws_session_token:str=None,
    config:Any=None,
    region_name:str=None,
   ) -> None:
    self._bucket_paths = bucket_paths


  def _bucket_path(self, bucket:str) -> str:
    try: return self._bucket_paths[bucket]
    except KeyError as e: raise S3MockError(f'mock bucket not found: {bucket!r}') from e


  def _open(self, bucket:str, key:str, mode:str) -> IO[Any]:
    bucket_path = self._bucket_path(bucket)
    key_dir = path_dir(key)
    assert not key_dir.startswith('/'), key_dir
    dir = path_join(bucket_path, key_dir)
    make_dirs(dir)
    path = path_join(bucket_path, key)
    try: return open(path, mode)
    except KeyError: raise S3MockError(f'mock key not found: {key!r}')


  def get_object(self, Bucket:str, Key:str) -> Dict[str,Any]:
    try: f =  self._open(Bucket, Key, 'rb')
    except FileNotFoundError: pass
    else:
      encoding:Optional[str] = None
      # TODO: metadata should be saved to disk and read, not inferred.
      # Guessing like this is not sufficiently generalized.
      if Key.endswith('.br'): encoding = 'br'
      elif Key.endswith('.gz'): encoding = 'gzip'
      with f:
        return {
          'Body': BytesIO(f.read()),
          'ResponseMetadata': {
            'HTTPHeaders': {
              'content-encoding': encoding
            }
          }
        }
    # Not found.
    error_response = {
      'Error': {
        'Code': 'NoSuchKey',
        'Key': Key,
        'Message': 'The specified key does not exist.'},
        'ResponseMetadata': {
          'HTTPHeaders': {},
          'HTTPStatusCode': 404,
        }
    }
    raise ClientError(error_response=error_response, operation_name='GetObject')


  def list_objects_v2(self, Bucket:str) -> Dict[str,Any]:
    path = self._bucket_path(Bucket)
    assert not path.endswith('/')
    l = len(path) + 1 # Eat the directory slash.
    contents = []
    for p in walk_files(path):
      key = p[l:]
      s = file_status(p, follow=True)
      assert s is not None
      contents.append({
        'Key': key,
        'Size': s.size,
        'LastModified': DateTime.fromtimestamp(s.mtime, tz=TimeZone.utc),
      })
    return {
      'KeyCount': len(contents),
      'Contents': contents,
    }


  def put_object(self, Bucket:str, Key:str, ContentType:str, ContentEncoding:str, Body:bytes) -> Dict[str,Any]:
    with self._open(Bucket, Key, 'wb') as f:
      f.write(Body)
    return {}


  def upload_file(self, path:str, Bucket:str, Key:str) -> Dict[str,Any]:
    with open(path, 'rb') as local_f:
      data = local_f.read()
    with self._open(Bucket, Key, 'wb') as f:
      f.write(data)
    return {}



def s3_client(session:Session, **kwargs:Any) -> S3Client:
  '''
  Create an S3MockClient if the `S3_MOCK_CLIENT` environment variable is set;
  otherwise return a normal s3 client.
  '''
  if 'S3_MOCK_CLIENT' in os.environ:
    bucket_paths:Dict[str, str] = {}
    for s in os.environ['S3_MOCK_CLIENT'].split():
      b, s, p = s.partition('=')
      if not b: raise ValueError(f'S3_MOCK_CLIENT word is not valid bucket name or `bucket=path` pair: {s!r}')
      if not s: p = b
      elif not p: raise ValueError(f'S3_MOCK_CLIENT bucket path is not valid: {s!r}')
      if b in bucket_paths: raise KeyError(f'S3_MOCK_CLIENT repeated bucket: {b!r}')
      bucket_paths[b] = p
    return S3MockClient(bucket_paths=bucket_paths, **kwargs)
  return session.client('s3', **kwargs) # type: ignore



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from types import GeneratorType
from typing import Callable, Generator, Iterable, Iterator, List, Optional, Tuple, TypeVar, Union

from .exceptions import OmitNode
from .iterable import iter_values


_T = TypeVar('_T')
_R = TypeVar('_R')

_Stack = Tuple[_T, ...]
_VisitResult = Union[_R, Generator[_R, None, None]]
_GetChildrenFn = Callable[[_T], Optional[Iterable[_T]]]
_TransformVisitor = Callable[[_T, _Stack, List[_R]], _VisitResult]


def transform_tree(root:_T, get_children:_GetChildrenFn, visit:_TransformVisitor) -> _R:
  '''
  `transform_tree` visits nodes, leaves-first, with the `visit` function,
  thereby generating a transformed tree.
  The `visit` function takes three parameters:
  * node: the current node.
  * stack: the stack of parent nodes.
  * results: the transformed children.

  `visit` should either:
  * return a single result node;
  * return a generator of results, which are flattened with those of the node's siblings;
  * raise OmitNode.
  '''
  res = _transform_tree(root, get_children, visit, ())
  if isinstance(res, GeneratorType): raise ValueError(res)
  else: return res # type: ignore


def _transform_tree(node:_T, get_children:_GetChildrenFn, visit:_TransformVisitor, stack:_Stack) -> _VisitResult:
  results: List[_R] = [] # type: ignore
  children = get_children(node)
  if children:
    child_stack = (*stack, node)
    for child in children:
      try: r = _transform_tree(child, get_children, visit, child_stack)
      except OmitNode: continue
      if isinstance(r, GeneratorType):
        results.extend(r)
      else:
        results.append(r)
  return visit(node, stack, results)


def visit_tree_preorder(obj:_T, get_values:Callable[[_T],Iterable[_T]]=iter_values) -> Iterator[_T]:
  yield obj
  for el in get_values(obj):
    yield from visit_tree_preorder(el, get_values)


def visit_tree_postorder(obj:_T, get_values:Callable[[_T],Iterator[_T]]=iter_values) -> Iterator[_T]:
  for el in get_values(obj):
    yield from visit_tree_postorder(el, get_values)
  yield obj



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Callable, Iterable, Iterator, List, TypeVar
from .default import Raise, RaiseOr


_T = TypeVar('_T')


class Buffer(Iterator[_T]):
  '''
  Iterable object that buffers an iterable.
  Call push() to push an item into the buffer;
  this will be returned on the subsequent call to next().
  '''


  def __init__(self, iterable: Iterable[_T]) -> None:
    self.iterator = iter(iterable)
    self.buffer: List[_T] = []


  def __repr__(self) -> str:
    return 'IterBuffer({!r}, buffer={!r})'.format(self.iterator, self.buffer)


  def __bool__(self) -> bool:
    try: self.peek()
    except StopIteration: return False
    else: return True


  def __iter__(self) -> Iterator[_T]: return self


  def __next__(self) -> _T:
    try: return self.buffer.pop()
    except IndexError: pass
    return next(self.iterator)


  def push(self, el:_T) -> None:
    self.buffer.append(el)


  def peek(self, default: RaiseOr[_T]=Raise._) -> _T:
    try: return self.buffer[-1]
    except IndexError: pass
    try: el = next(self.iterator)
    except StopIteration:
      if isinstance(default, Raise): raise
      else: return default
    self.buffer.append(el)
    return el


  def take_while(self, pred: Callable[[_T], bool]) -> Iterator[_T]:
    for el in self:
      if pred(el):
        yield el
      else:
        self.buffer.append(el)
        break


  def drop_while(self, pred: Callable[[_T], bool]) -> None:
    for el in self:
      if not pred(el):
        self.buffer.append(el)
        break


  def peek_while(self, pred: Callable[[_T], bool]) -> List[_T]:
    els = list(self.take_while(pred))
    self.buffer.extend(reversed(els))
    return els


  def take(self, count: int, short=False, default: RaiseOr[_T]=Raise._) -> List[_T]:
    els = []
    for _ in range(count):
      try: els.append(next(self))
      except StopIteration:
        if short: break
        if isinstance(default, Raise): raise
        els.append(default)
    return els


  def peeks(self, count: int, short=False, default: RaiseOr[_T]=Raise._) -> List[_T]:
    if 0 < count <= len(self.buffer):
      return list(reversed(self.buffer[-count:]))
    els = []
    for _ in range(count):
      try: els.append(next(self))
      except StopIteration:
        if short: break
        if isinstance(default, Raise): raise
        els.append(default)
    self.buffer.extend(reversed(els))
    return els


  def peek_all(self) -> List[_T]:
    return self.peeks(count=1<<63, short=True)


  def expect(self, pred: Callable[[_T], bool]) -> _T:
    el = next(self)
    if pred(el): return el
    raise ValueError(el)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import (Any, Callable, Dict, Hashable, Iterable, List, Mapping, MutableMapping, NamedTuple, Set, Tuple, TypeVar,
  Union)


_K = TypeVar('_K', bound=Hashable)
_V = TypeVar('_V')


class ConflictingValues(KeyError): pass


class KeyExistingIncoming(NamedTuple):
  'Triple of (key, existing, incoming) for reporting conflicting values.'
  key:Any
  existing:Any
  incoming:Any


def dict_discard(d:MutableMapping[_K,_V], k:_K) -> None:
  'Discard the element at `k` if it exists.'
  try: del d[k]
  except KeyError: pass


def dict_put(d:MutableMapping[_K,_V], k: _K, v: _V) -> MutableMapping[_K,_V]:
  '''
  Put a new key and value in the dictionary, or raise err KeyError if the key already exists.
  Returns the dictionary.
  '''
  if k in d: raise KeyError(k)
  d[k] = v
  return d


def idemput(d:MutableMapping[_K,_V], k:_K, v:_V) -> MutableMapping[_K,_V]:
  '''
  Put a new key and value in the dictionary;
  raise err KeyError if the key already exists and the existing value is not equal to the incoming one.
  '''
  try: existing = d[k]
  except KeyError: pass
  else:
    if v != existing: raise ConflictingValues(KeyExistingIncoming(k, existing, v))
  d[k] = v
  return d


def dict_list_append(d: Dict[_K, List[_V]], k: _K, v: _V) -> Dict[_K, List[_V]]:
  '''
  Append a value to the list stored under the specified key in the dictionary.
  If the key is not present, an empty list is first inserted.
  Returns the dictionary.
  Note: equivalent behavior can also be achieved with a `defaultdict(list)`.
  '''
  d.setdefault(k, []).append(v)
  return d


def dict_list_extend(d: Dict[_K, List[_V]], k: _K, v: Iterable[_V]) -> Dict[_K, List[_V]]:
  '''
  Extend a value on the list stored under the specified key in the dictionary.
  If the key is not present, an empty list is first inserted.
  Returns the dictionary.
  Note: equivalent behavior can also be achieved with a `defaultdict(list)`.
  '''
  d.setdefault(k, []).extend(v)
  return d


def dict_set_defaults(d: MutableMapping[_K, _V], defaults: Union[Mapping[_K, _V], Iterable[Tuple[_K, _V]]]) -> MutableMapping[_K, _V]:
  '''
  Call setdefault on the dictionary for each item in `defaults`,
  which can be either dictionary-like object implementing `items()` or a sequence of pairs.
  Returns the dictionary.
  '''
  it: Iterable[Tuple[_K, _V]]
  if isinstance(defaults, Mapping):
    it = defaults.items()
  else:
    it = defaults
  for k, v in it:
    d.setdefault(k, v)
  return d


_VH = TypeVar('_VH', bound=Hashable)
_S = Set[_VH]
_I = Iterable[_VH]

def dict_update_sets(d:MutableMapping[_K,_S], update:Union[Mapping[_K, _I],Iterable[Tuple[_K,_I]]]) -> MutableMapping[_K, _S]:
  it:Iterable[Tuple[_K,_I]]
  if isinstance(update, Mapping):
    it = update.items()
  else:
    it = update
  for k, i in it:
    try: s = d[k]
    except KeyError:
      s = set()
      d[k] = s
    s.update(i)
  return d


def dict_fan_by_key_pred(d: Mapping[_K, _V], pred: Callable[[_K], bool]) -> Tuple[Dict[_K, _V], Dict[_K, _V]]:
  'Fan out `d` into a pair of dictionaries by applying `pred` to each key in `d`.'
  fan: Tuple[Dict[_K, _V], Dict[_K, _V]] = ({}, {})
  for k, v in d.items():
    if pred(k):
      fan[1][k] = v
    else:
      fan[0][k] = v
  return fan


class DefaultByKeyDict(Dict[_K,_V]):
  '''
  Subclass of Dict, similar to DefaultDict.
  When a key is missing, default_factory is called with the key as the sole argument.
  '''
  def __init__(self, default_factory:Callable[[_K], _V], *args, **kwargs):
    self.default_factory = default_factory
    super().__init__(*args, **kwargs)

  def __missing__(self, key):
    val = self.default_factory(key)
    self[key] = val
    return val



#!/usr/bin/env python3

from math import sqrt
from sys import argv
from time import perf_counter
from typing import List

from pithy.task import DEVNULL, communicate, launch


def main() -> None:
  if len(argv) < 3: exit(f'error: usage: time-runs NUM_RUNS CMD ...')

  _, num_runs_str, *cmd = argv
  try: num_runs = int(num_runs_str)
  except ValueError: exit(f'error: first argument must be an integer; received {argv[1]!r}.')
  if num_runs < 1: exit(f'error: first argument must be positive: received {num_runs}.')

  tot_times:List[float] = []
  wait_times:List[float] = []
  for i in range(num_runs):
    pre_launch_time = perf_counter()
    _cmd, proc, _input = launch(cmd, err=DEVNULL, out=DEVNULL)
    post_launch_time = perf_counter()
    status, _, _ = communicate(proc)
    end_time = perf_counter()
    tot_time = end_time - pre_launch_time
    wait_time = end_time - post_launch_time
    tot_times.append(tot_time)
    wait_times.append(wait_time)
    print(f'{i:3d}: total: {tot_time:8.3f}; wait: {wait_time:8.3f}; exit status: {status}.')

  calc_stats(tot_times, label='total')
  calc_stats(wait_times, label='wait')


def calc_stats(times:List[float], label:str) -> None:
  l = len(times)
  avg = sum(times) / l
  std_dev = sqrt(sum((t-avg)**2 for t in times))
  print(f'{label:5}: avg: {avg:7.3f}; std-dev: {std_dev:6.3f}.')



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from sys import stderr
from time import time as now
from .typing import OptTypeBaseExc, OptBaseExc, OptTraceback


class Clock:

  def __init__(self, name:str) -> None:
    self.name = name
    self.times = [(now(), 'start')]


  def __enter__(self) -> 'Clock': return self


  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self.total()


  def split(self, label:str) -> None:
    t = now()
    print(f'clock: {self.name}: {label}: {t - self.times[-1][0]:.04f}', file=stderr)
    self.times.append((t, label))


  def total(self) -> None:
    t = now()
    print(f'clock: {self.name}: TOTAL: {t - self.times[0][0]:.04f}', file=stderr)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

# TODO: handle leading UTF8-BOM.

import csv
from csv import QUOTE_ALL, QUOTE_MINIMAL, QUOTE_NONNUMERIC, QUOTE_NONE, Dialect
from sys import stdout
from typing import Any, Callable, ContextManager, Iterable, Iterator, Optional, Sequence, TextIO, Type, Union

from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


def write_csv(f:TextIO, *, quoting:int=QUOTE_MINIMAL, header:Optional[Sequence[str]], rows:Iterable[Sequence]) -> None:
  w = csv.writer(f, quoting=quoting)
  if header is not None: w.writerow(header)
  w.writerows(rows)


def out_csv(*, quoting:int=QUOTE_MINIMAL, header:Optional[Sequence[str]], rows:Iterable[Sequence]) -> None:
  write_csv(f=stdout, quoting=quoting, header=header, rows=rows)


def load_csv(file: TextIO,
 dialect:Union[str,Dialect,Type[Dialect]]='excel',
 delimiter:Optional[str]=None,
 doublequote:Optional[bool]=None,
 escapechar:Optional[str]=None,
 quotechar:Optional[str]=None,
 quoting:int=QUOTE_MINIMAL,
 skipinitialspace:Optional[bool]=None,
 strict:Optional[bool]=None,
 row_type:type=None,
 col_conv:Iterable[Callable]=(),
 header:Union[None, bool, Sequence[str]]=None) -> 'CSVFileReader':

  return CSVFileReader(
    file=file,
    dialect=dialect,
    delimiter=delimiter,
    doublequote=doublequote,
    escapechar=escapechar,
    quotechar=quotechar,
    quoting=quoting,
    skipinitialspace=skipinitialspace,
    strict=strict,
    row_type=row_type,
    col_conv=col_conv,
    header=header)


class CSVFileReader(Iterable, ContextManager):

  def __init__(self, file: TextIO,
   dialect:Union[str,Dialect,Type[Dialect]]='excel',
   delimiter:Optional[str]=None,
   doublequote:Optional[bool]=None,
   escapechar:Optional[str]=None,
   quotechar:Optional[str]=None,
   quoting:Optional[int]=None,
   skipinitialspace:Optional[bool]=None,
   strict:Optional[bool]=None,
   row_type:type=None,
   col_conv:Iterable[Callable]=(),
   header:Union[None, bool, Sequence[str]]=None) -> None:

    opts = { k : v for (k, v) in [
      ('delimiter', delimiter),
      ('doublequote', doublequote),
      ('escapechar', escapechar),
      ('quotechar', quotechar),
      ('quoting', quoting),
      ('skipinitialspace', skipinitialspace),
      ('strict', strict),
      ] if v is not None }

    self._reader = csv.reader(file, dialect, **opts)
    self.file = file
    self.row_type = row_type
    self.col_conv = col_conv

    if header is None or isinstance(header, bool):
      if header: next(self._reader) # simply discard.
    else: # match expected against actual.
      row = next(self._reader)
      list_header = list(header)
      if row != list_header:
        raise ValueError(f'load_csv expected header:\n{list_header}\nreceived:\n{row}')


  def __iter__(self) -> Iterator[Any]:
    if self.col_conv:
      if self.row_type is None:
        return ([conv(col) for conv, col in zip(self.col_conv, row)] for row in self._reader)
      else:
        return (self.row_type(*(conv(col) for conv, col in zip(self.col_conv, row))) for row in self._reader)
    else:
      if self.row_type is None:
        return self._reader
      else:
        return (self.row_type(*row) for row in self._reader)


  def __enter__(self) -> 'CSVFileReader':
    return self


  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self.file.close()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from os import O_NONBLOCK, O_RDONLY, close as os_close, open as os_open, read as os_read
from pprint import pprint
from shlex import quote as sh_quote
from string import Template as _Template
from sys import stderr, stdin, stdout
from typing import Any, Callable, ContextManager, Iterable, Iterator, Sized, TextIO, TypeVar, Union, cast

from .desc import errD, outD, writeD
from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


_T = TypeVar('_T')

# basic printing.

def writeZ(file: TextIO, *items: Any, sep='', end='', flush=False) -> None:
  "Write `items` to file; default sep='', end=''."
  print(*items, sep=sep, end=end, file=file, flush=flush)

def writeS(file: TextIO, *items: Any, sep='', flush=False) -> None:
  "Write `items` to file; sep='', end=' '."
  print(*items, sep=sep, end=' ', file=file, flush=flush)

def writeSZ(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep=' ', end=''."
  print(*items, sep=' ', end='', file=file, flush=flush)

def writeSS(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep=' ', end=''."
  print(*items, sep=' ', end=' ', file=file, flush=flush)

def writeL(file: TextIO, *items: Any, sep='', flush=False) -> None:
  "Write `items` to file; sep='', end='\\n'."
  print(*items, sep=sep, end='\n', file=file, flush=flush)

def writeN(file: TextIO, *items: Any, sep='', flush=False) -> None:
  "Write `items` to file; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=sep, end=end, file=file, flush=flush)

def writeSL(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep=' ', end='\\n'."
  print(*items, sep=' ', end='\n', file=file, flush=flush)

def writeSN(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=' ', end=end, file=file, flush=flush)

def writeLL(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep='\\n', end='\\n'."
  print(*items, sep='\n', end='\n', file=file, flush=flush)

def writeLSSL(file: TextIO, *items: Any, flush=False) -> None:
  "Write `items` to file; sep='\\n  ', end='\\n'."
  print(*items, sep='\n  ', end='\n', file=file, flush=flush)


# templated format printing.

def writeTF(file:TextIO, template_fmt:str, *items:Any, flush=False, **keyed_items:Any) -> None:
  """
  Expand the format string with keyed_items, then format the string; end=''.
  Useful for constructing dynamic format strings.
  """
  fmt = _Template(template_fmt).substitute(**keyed_items)
  print(fmt.format(*items, **keyed_items, end='', file=file, flush=flush))


def writeTFL(file:TextIO, template_fmt:str, *items:Any, flush=False, **keyed_items:Any) -> None:
  """
  Expand the format string template with keyed_items, then format the string; end='\\n'
  Useful for constructing dynamic format strings.
  """
  fmt = _Template(template_fmt).substitute(**keyed_items)
  print(fmt.format(*items, **keyed_items, file=file, flush=flush))


def writeP(file: TextIO, *items: Any, label=None, indent=2, **opts: Any) -> None:
  'Pretty print to file.'
  if label is not None:
    file.write(label)
    file.write (': ')
  for item in items:
    pprint(item, stream=file, indent=indent, **opts)


# std out.

def outZ(*items: Any, sep='', end='', flush=False) -> None:
  "Write `items` to std out; sep='', end=''."
  print(*items, sep=sep, end=end, flush=flush)

def outS(*items: Any, sep='', flush=False) -> None:
  "Write `items` to std out; sep='', end=' '."
  print(*items, sep=sep, end=' ', flush=flush)

def outSZ(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep=' ', end=''."
  print(*items, sep=' ', end='', flush=flush)

def outSS(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep=' ', end=' '."
  print(*items, end=' ', flush=flush)

def outL(*items: Any, sep='', flush=False) -> None:
  "Write `items` to std out; sep='', end='\\n'."
  print(*items, sep=sep, flush=flush)

def outN(*items: Any, sep='', flush=False) -> None:
  "Write `items` to std out; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=sep, end=end, flush=flush)

def outSL(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep=' ', end='\\n'."
  print(*items, flush=flush)

def outSN(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=' ', end=end, flush=flush)

def outLL(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep='\\n', end='\\n'."
  print(*items, sep='\n', flush=flush)

def outLSSL(*items: Any, flush=False) -> None:
  "Write `items` to std out; sep='\\n  ', end='\\n'."
  print(*items, sep='\n  ', flush=flush)

def outP(*items: Any, label=None, flush=False, **opts: Any) -> None:
  'Pretty print to std out.'
  writeP(stdout, *items, label=label, **opts)


# std err.

def errZ(*items: Any, sep='', end='', flush=False) -> None:
  "Write items to std err; default sep='', end=''."
  print(*items, sep=sep, end=end, file=stderr, flush=flush)

def errS(*items: Any, sep='', flush=False) -> None:
  "Write items to std err; sep='', end=' '."
  print(*items, sep=sep, end=' ', file=stderr, flush=flush)

def errSZ(*items: Any, flush=False) -> None:
  "Write items to std err; sep=' ', end=''."
  print(*items, sep=' ', end='', file=stderr, flush=flush)

def errSS(*items: Any, flush=False) -> None:
  "Write items to std err; sep=' ', end=''."
  print(*items, sep=' ', end=' ', file=stderr, flush=flush)

def errL(*items: Any, sep='', flush=False) -> None:
  "Write items to std err; sep='', end='\\n'."
  print(*items, sep=sep, end='\n', file=stderr, flush=flush)

def errN(*items: Any, sep='', flush=False) -> None:
  "Write `items` to std err; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=sep, end=end, file=stderr, flush=flush)

def errSL(*items: Any, flush=False) -> None:
  "Write items to std err; sep=' ', end='\\n'."
  print(*items, sep=' ', end='\n', file=stderr, flush=flush)

def errSN(*items: Any, flush=False) -> None:
  "Write `items` to std err; sep='', end=('' if items and items[-1].endswith('\\n') else '\\n')."
  end = ('' if items and items[-1].endswith('\n') else '\n')
  print(*items, sep=' ', end=end, file=stderr, flush=flush)

def errLL(*items: Any, flush=False) -> None:
  "Write items to std err; sep='\\n', end='\\n'."
  print(*items, sep='\n', end='\n', file=stderr, flush=flush)

def errLSSL(*items: Any, flush=False) -> None:
  "Write items to std err; sep='\\n  ', end='\\n'."
  print(*items, sep='\n  ', end='\n', file=stderr, flush=flush)

def errP(*items: Any, label=None, **opts) -> None:
  'Pretty print to std err.'
  writeP(stderr, *items, label=label, **opts)


def err_progress(iterable: Iterable[_T], label='progress', suffix='', final_suffix='', frequency:Union[float, int]=0.1, limit=0) -> Iterator[_T]:
  '''
  For interactive terminals, return a generator that yields the elements of `iterable`
  and displays a progress indicator on std err.
  '''
  assert frequency >= 0
  if not frequency or not stderr.isatty():
    return iter(iterable)

  ERASE_LINE = '\x1b[2K'

  if label is None:
    label = str(iterable)
  pre = f'{ERASE_LINE}\r◊ {label}: '
  post = (suffix and ' ' + suffix) + '…'
  final = f' {final_suffix}.' if final_suffix else '.'

  total = ''
  width = 0
  try: l = len(cast(Sized, iterable))
  except TypeError: pass
  else:
    ls = f'{l:,}'
    width = len(ls)
    total = '/' + ls

  if isinstance(frequency, float):
    from time import time
    def err_progress_gen() -> Iterator[_T]:
      prev_t = time()
      step = 1
      next_i = step
      i = -1
      for i, el in enumerate(iterable):
        if limit and i == limit:
          i -= 1
          break
        if i == next_i:
          print(f'{pre}{i:{width},}{total}{post}', end='', file=stderr, flush=True)
          t = time()
          d = t - prev_t
          step = max(1, int(step * frequency / d))
          prev_t = t
          next_i = i + step
        yield el
      print(f'{pre}{i+1:{width},}{total}{final}', file=stderr)

  else: # frequency is an int.
    def err_progress_gen() -> Iterator[_T]: # type: ignore # Spurious "Statement is unreachable".
      for i, el in enumerate(iterable):
        if limit and i == limit:
          i -= 1
          break
        if i % frequency == 0:
          print(pre + str(i) + post, end='', file=stderr, flush=True)
        yield el
      print(pre + str(i) + final, file=stderr)

  return err_progress_gen()


# convenience read/write.


def read_from_path(path: str, default: str=None) -> str:
  'Read all text from file at `path`.'
  try:
    with open(path) as f:
      return f.read()
  except (FileNotFoundError, IsADirectoryError):
    if default is None: raise
    return default


def read_line_from_path(path: str, line_index=0, keep_end=False, default: str=None) -> str:
  'Read a single line of text from file at `path`.'
  try:
    with open(path) as f:
      for i, line in enumerate(f):
        if i == line_index:
          return line if keep_end else line.rstrip('\n')
      if default is None: raise IndexError(line_index)
      return default
  except (FileNotFoundError, IsADirectoryError, PermissionError, UnicodeDecodeError):
    if default is None: raise
    return default


def write_to_path(path:str, text:Union[str,bytes,bytearray]) -> None:
  'Writes `string` to file at `path`.'
  if isinstance(text, str):
    with open(path, 'w') as f: f.write(text)
  else:
    with open(path, 'wb') as bf: bf.write(text)


# Opener utility.

def mk_opener(flags:int, mode=0o777, dir_fd:int=None) -> Callable[[str, int], int]:
  def _opener(path:str, _flags:int, mode=mode, dir_fd=dir_fd) -> int: return os_open(path,_flags&flags)
  return _opener

nonblock_opener = mk_opener(O_NONBLOCK)


# Nonblocking tools.

class AsyncLineReader(ContextManager):
  '''
  A file-like object for reading asynchronously from a file descriptor.
  '''

  def __init__(self, path:str) -> None:
    self.fd = os_open(path, O_RDONLY|O_NONBLOCK) # TODO: accept raw descriptor.
    self.buffer = bytearray()

  def __del__(self) -> None:
    self.close()

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self.close()

  def close(self) -> None:
    if self.fd >= 0:
      os_close(self.fd)
      self.fd = -1

  def readline(self) -> str:
    '''
    Attempt to return a complete line from the input stream.
    If there is not enough data available, return ''.
    '''
    line:Union[bytes,bytearray]

    # The buffer might already contain a complete line.
    buffer_line_end = self.buffer.find(0x0a) + 1 # 0 if no newline found.
    if buffer_line_end: # Buffer already contains a complete line.
      line = self.buffer[:buffer_line_end]
      del self.buffer[buffer_line_end:]
      return line.decode()

    # Read from the file descriptor until it returns nothing or we have a complete line.
    while True:
      try: data = os_read(self.fd, 4096)
      except BlockingIOError: return ''
      if not data: return ''
      line_end = data.find(0x0a) + 1 # 0 if no newline found.
      if not line_end: # No newline.
        self.buffer.extend(data)
        continue
      if self.buffer: # Have previous data.
        line_end += len(self.buffer)
        self.buffer.extend(data)
        line = self.buffer[:line_end]
        del self.buffer[:line_end]
      else: # No previous data.
        line = data[:line_end]
        self.buffer.extend(data[line_end:])
      return line.decode()


# misc.

def clip_newlines(iterable: Iterable[str]) -> Iterable[str]:
  for line in iterable:
    yield line.rstrip('\n')


def confirm(question:str) -> bool:
  from .term import CBreakMode
  print(f'{question}? press "y" to confirm: ', end='', flush=True)
  with CBreakMode(): response = stdin.read(1)
  print(response)
  return (response == 'y')


def confirm_or_exit(question:str) -> None:
  if not confirm(question): exit(1)


def assert_eq(a: Any, b: Any):
  if a != b:
    raise AssertionError(f'not equal:\n  {a!r}\n  {b!r}')


def shell_cmd_str(cmd:Iterable[str]) -> str:
  return ' '.join(sh_quote(word) for word in cmd)


def tee_to_err(iterable:Iterable[_T], label:str = 'tee_to_err') -> Iterator[_T]:
  for el in iterable:
    errL(label, ': ', repr(el))
    yield el



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from . import Syntax
from ..dispatch import dispatched
from ..io import errL
from ..tree import transform_tree, OmitNode
from dataclasses import dataclass
from typing import Iterable, List, Optional, Tuple
import re

from docutils import frontend as _frontend
from docutils import nodes as _nodes
from docutils.nodes import Node as Node, Text
from docutils.utils import new_document as _new_document
from docutils.parsers.rst import Parser as _RstParser


def parse_rst(path:str, text:str) -> Syntax:
  parser = _RstParser()
  settings = _frontend.OptionParser(components=(_RstParser,)).get_default_values()
  document = _new_document(path, settings=settings)
  parser.parse(text, document)
  ctx = _Ctx(path=path, text=text, lines=text.splitlines(keepends=True))
  return transform_tree(document, _get_children, ctx.visit)


@dataclass
class Ref:
  text: Syntax
  target: Optional[Syntax]

  def __iter__(self):
    yield self.text
    if self.target: yield self.target


def _get_children(node:Node) -> Iterable[Node]:
  return node.children # type: ignore


@dataclass
class _Ctx:
  path: str
  text: str
  lines: List[str]
  line: int = 0
  col: int = 0

  @property
  def curr_line_text(self) -> str: return self.lines[self.line]

  def match_text(self, text:str, skip_leading:bool, label:str, raises=False) -> Syntax.Pos:
    'Match a single line of text, advancing self.line and self.col.'
    if skip_leading:
      while rst_syntax_re.fullmatch(self.curr_line_text, self.col):
        #errL(f'SKIPPED: {self.curr_line_text[self.col:]!r}')
        self.line += 1
        self.col = 0
    line = self.line
    col = self.curr_line_text.find(text, self.col)
    if col == -1:
      if raises: raise ValueError(text)
      errL(Syntax.diagnostic(self.path, line=self.line, col=self.col,
        msg=f'warning: {label} not matched: {text!r}\n  {self.curr_line_text!r}'))
      col = self.col # better than nothing.
      end_col = -1
    else: # matched.
      end_col = col+len(text)
      if text.endswith('\n'):
        self.line +=1
        self.col = 0
      else:
        self.col = end_col
      #errL(f'MATCHED: {self.line}:{self.col}: {text!r}')
    return Syntax.Pos(line=line, col=col, end_line=self.line, end_col=end_col)


  @dispatched
  def visit(self, node:Node, stack:Tuple[Node, ...], children:List[Node]) -> Syntax:
    'Default visitor.'
    pos = Syntax.Pos(line=(-1 if node.line is None else node.line-1), enclosed=children)
    return Syntax(path=self.path, pos=pos, kind=_kind_for(node), content=children)


  @dispatched # type: ignore
  def visit(self, node:Text, stack:Tuple[Node, ...], children:List[Node]) -> Syntax:
    'Text visitor. Determines line/col position post-hoc, which docutils does not provide.'
    assert node.line is None # Text never has line number.
    text = node.astext()
    text_lines = text.splitlines(keepends=True)
    if len(text_lines) == 1:
      pos = self.match_text(text, skip_leading=True, label='text')
      return Syntax(path=self.path, pos=pos, kind='text', content=text)
    # multiline text blocks.
    children = []
    for i, text_line in enumerate(text_lines):
      is_lead = not i
      label = 'lead multiline text' if is_lead else 'tail multiline text'
      pos = self.match_text(text_line, skip_leading=is_lead, label=label)
      children.append(Syntax(path=self.path, pos=pos, kind='text', content=text_line))
    pos = Syntax.Pos(enclosed=children)
    return Syntax(path=self.path, pos=pos, kind='lines', content=tuple(children))


  @dispatched # type: ignore
  def visit(self, node:_nodes.reference, stack:Tuple[Node, ...], children:List[Node]) -> Syntax:
    assert len(children) == 1
    text = children[0]
    assert isinstance(text, Syntax)
    attrs = node.attributes
    uri = attrs.get('refuri')
    target = None
    if uri:
      try: pos = self.match_text(uri, skip_leading=True, label='ref uri', raises=True)
      except ValueError: pass # might have been extracted from the text.
      else: target = Syntax(path=self.path, pos=pos, kind='target', content=uri)
    pos = Syntax.Pos(enclosed=(text, target))
    content = Ref(text=text, target=target)
    return Syntax(path=self.path, pos=pos, kind='ref', content=content)


  @dispatched # type: ignore
  def visit(self, node:_nodes.target, stack:Tuple[Node, ...], children:List[Node]) -> None:
    raise OmitNode


rst_syntax_re = re.compile(r'[~=\s\-_<>`:]*') # matches lines that are only markup syntax and not content.


def _kind_for(node:Node) -> str: return type(node).__name__
